__author__ = 'Georgios Rizos (georgerizos@iti.gr)'

from reveal_user_classification.datautil.asu_read_data import read_adjacency_matrix
# from reveal_user_classification.datautil.snow_datautil.snow_read_data import read_adjacency_matrix
# from reveal_user_classification.datautil.insight_datautil.insight_read_data import read_adjacency_matrix
from reveal_user_classification.eps_randomwalk.benchmarks.benchmark_design import similarity_slice_benchmark,\
    aggregate_measurements, print_stats


########################################################################################################################
# Configure experiment.
########################################################################################################################
# Select dataset
DATASET = "YouTube_cc"  # Choices are: BlogCatalog, Flickr, YouTube_cc.

# Define approximate PageRank method parameters.
RHO_EFFECTIVE = 0.1
EPSILON = 0.00001

# Define laziness factor for the lazy PageRank.
LAZINESS = 0.5

NUMBER_OF_TRIALS = 2

########################################################################################################################
# Read data.
########################################################################################################################
# Define data path.
EDGE_LIST_PATH = get_raw_datasets_path() + "/ASU/" + DATASET + "/edges.csv"
adjacency_matrix = read_adjacency_matrix(EDGE_LIST_PATH, ',')
number_of_nodes = adjacency_matrix.shape[0]

profile_individual_stat_folder = get_memory_path() + "/ASU/" + DATASET + "/profile/individual"
profile_aggregate_stat_folder = get_memory_path() + "/ASU/" + DATASET + "/profile/aggregate"
profile_output_stat_folder = get_memory_path() + "/ASU/" + DATASET + "/profile/output"

########################################################################################################################
# Perform experiment.
########################################################################################################################
# profiler_stats = similarity_slice_benchmark(number_of_trials=NUMBER_OF_TRIALS,
#                                             adjacency_matrix=adjacency_matrix,
#                                             rho_effective=RHO_EFFECTIVE,
#                                             epsilon=EPSILON,
#                                             laziness_factor=LAZINESS,
#                                             profile_individual_stat_folder=profile_individual_stat_folder)

# aggregate_measurements(number_of_trials=NUMBER_OF_TRIALS,
#                        adjacency_matrix=adjacency_matrix,
#                        profile_individual_stat_folder=profile_individual_stat_folder,
#                        profile_aggregate_stat_folder=profile_aggregate_stat_folder)

print_stats(number_of_trials=NUMBER_OF_TRIALS,
            profile_aggregate_stat_folder=profile_aggregate_stat_folder,
            profile_output_stat_folder=profile_output_stat_folder)



__author__ = 'Georgios Rizos (georgerizos@iti.gr)'

import numpy as np
from sklearn.kernel_approximation import AdditiveChi2Sampler
from sklearn.preprocessing import normalize, scale


def normalize_community_features(features):
    """
    This performs TF-IDF-like normalization of community embedding features.

    Introduced in: Tang, L., Wang, X., Liu, H., & Wang, L. (2010, July).
                   A multi-resolution approach to learning with overlapping communities.
                   In Proceedings of the First Workshop on Social Media Analytics (pp. 14-22). ACM.

    Input:  - X in R^(nxC_n): The community indicator matrix.

    Output: - X_norm in R^(nxC_n): The tf-idf + row normalized community indicator matrix.
    """
    # Calculate inverse document frequency.
    features = normalize_columns(features)

    # Normalize each row of term frequencies to 1
    features = normalize_rows(features)

    return features


def normalize_rows(features):
    """
    This performs row normalization to 1 of community embedding features.

    Input:  - X in R^(nxC_n): The community indicator matrix.

    Output: - X_norm in R^(nxC_n): The row normalized community indicator matrix.
    """
    # Normalize each row of term frequencies to 1
    features = features.tocsr()
    features = normalize(features, norm="l2")
    # for i in range(features.shape[0]):
    #     term_frequency = features.getrow(i).data
    #     if term_frequency.size > 0:
    #         features.data[features.indptr[i]: features.indptr[i + 1]] =\
    #             features.data[features.indptr[i]: features.indptr[i + 1]]/np.sqrt(np.sum(np.power(term_frequency, 2)))

    return features


def normalize_columns(features):
    """
    This performs column normalization of community embedding features.

    Input:  - X in R^(nxC_n): The community indicator matrix.

    Output: - X_norm in R^(nxC_n): The tf-idf + row normalized community indicator matrix.
    """
    # Calculate inverse document frequency.
    features = features.tocsc()
    for j in range(features.shape[1]):
        document_frequency = features.getcol(j).data.size
        if document_frequency > 1:
            features.data[features.indptr[j]: features.indptr[j + 1]] =\
                features.data[features.indptr[j]: features.indptr[j + 1]]/np.sqrt(np.log(document_frequency))

    features = features.tocsr()

    return features


def scale_columns(features):
    features = scale(features, with_mean=False)

    return features


def chi_squared_projection(features):
    chi2_feature = AdditiveChi2Sampler()
    X_transformed = chi2_feature.fit_transform(features)
    X_transformed = X_transformed.tocsr()

    return X_transformed



__author__ = 'Georgios Rizos (georgerizos@iti.gr)'

import os
import inspect
import multiprocessing
try:
    import cPickle as pickle
except ImportError:
    import pickle

import reveal_graph_embedding


########################################################################################################################
# Configure path related functions.
########################################################################################################################
def get_package_path():
    return os.path.dirname(inspect.getfile(reveal_graph_embedding))


########################################################################################################################
# Configure optimization related functions.
########################################################################################################################
def get_threads_number():
    """
    Automatically determine the number of cores. If that fails, the number defaults to a manual setting.
    """
    try:
        cores_number = multiprocessing.cpu_count()
        return cores_number
    except NotImplementedError:
        cores_number = 8
        return cores_number


def get_file_row_generator(file_path, separator, encoding=None):
    """
    Reads an separated value file row by row.

    Inputs: - file_path: The path of the separated value format file.
            - separator: The delimiter among values (e.g. ",", "\t", " ")
            - encoding: The encoding used in the stored text.

    Yields: - words: A list of strings corresponding to each of the file's rows.
    """
    with open(file_path, encoding=encoding) as file_object:
        for line in file_object:
            words = line.strip().split(separator)
            yield words


def store_pickle(file_path, data):
    """
    Pickle some data to a given path.

    Inputs: - file_path: Target file path.
            - data: The python object to be serialized via pickle.
    """
    pkl_file = open(file_path, 'wb')
    pickle.dump(data, pkl_file)
    pkl_file.close()


def load_pickle(file_path):
    """
    Unpickle some data from a given path.

    Input:  - file_path: Target file path.

    Output: - data: The python object that was serialized and stored in disk.
    """
    pkl_file = open(file_path, 'rb')
    data = pickle.load(pkl_file)
    pkl_file.close()
    return data



__author__ = 'Georgios Rizos (georgerizos@iti.gr)'
# cython: profile=True

import numpy as np
cimport numpy as np
cimport cython
import scipy.sparse as sparse


def get_natural_random_walk_matrix(adjacency_matrix):
    """
    Returns the natural random walk transition probability matrix given the adjacency matrix.

    Input:  - A: A sparse matrix that contains the adjacency matrix of the graph.

    Output: - W: A sparse matrix that contains the natural random walk transition probability matrix.
    """
    # Turn to sparse.csr_matrix format for faster row access.
    rw_transition = sparse.csr_matrix(adjacency_matrix, dtype=np.float64)

    # Sum along the two axes to get out-degree and in-degree, respectively
    out_degree = rw_transition.sum(axis=1)
    in_degree = rw_transition.sum(axis=0)

    # Form the inverse of the diagonal matrix containing the out-degree
    for i in np.arange(rw_transition.shape[0]):
        rw_transition.data[rw_transition.indptr[i]: rw_transition.indptr[i + 1]] =\
            rw_transition.data[rw_transition.indptr[i]: rw_transition.indptr[i + 1]]/out_degree[i]

    rw_transition.sort_indices()

    out_degree = np.array(out_degree).astype(np.float64).reshape(out_degree.size)
    in_degree = np.array(in_degree).astype(np.float64).reshape(in_degree.size)

    return rw_transition, out_degree, in_degree



__author__ = 'Georgios Rizos (georgerizos@iti.gr)'

import numpy as np
import numpy.linalg as npla
import scipy as sp
import scipy.sparse as spsp
import scipy.sparse.linalg as spla

from reveal_graph_embedding.eps_randomwalk.transition import get_natural_random_walk_matrix


def get_implicit_adjacency_matrices(rw_transition, rho=0.2):
    # Calculate random walk with restart and teleportation.
    rw_transition = rw_transition.tocoo()
    rw_transition_t = rw_transition.T.tocsr()
    rw_transition = rw_transition.tocsr()

    stationary_distribution = get_pagerank_with_teleportation(rw_transition,
                                                              rw_transition_t,
                                                              rho)

    # Calculate implicit combinatorial adjacency matrix.
    implicit_combinatorial_matrix, com_phi = get_implicit_combinatorial_adjacency_matrix(stationary_distribution,
                                                                                         rw_transition,
                                                                                         rw_transition_t)

    # Calculate implicit directed adjacency matrix.
    implicit_directed_matrix, dir_phi = get_implicit_directed_adjacency_matrix(stationary_distribution,
                                                                               rw_transition,
                                                                               rho)

    return implicit_combinatorial_matrix, com_phi, implicit_directed_matrix, dir_phi


def get_pagerank_with_teleportation(rw_transition, rw_transition_t, rho=0.2):
    number_of_nodes = rw_transition.shape[0]

    non_teleportation = 1-rho
    mv = lambda l, v: non_teleportation*l.dot(v) + (rho/number_of_nodes)*np.ones_like(v)
    teleport = lambda vec: mv(rw_transition_t, vec)

    rw_transition_operator = spla.LinearOperator(rw_transition.shape, matvec=teleport, dtype=np.float64)

    ####################################################################################################################
    # Form theta matrix
    ####################################################################################################################
    # Form stationary distribution diagonal matrix
    try:
        eigenvalue, stationary_distribution = spla.eigs(rw_transition_operator,
                                                        k=1,
                                                        which='LM',
                                                        return_eigenvectors=True)
    except spla.ArpackNoConvergence as e:
        print("ARPACK has not converged.")
        eigenvalue = e.eigenvalues
        stationary_distribution = e.eigenvectors

    return stationary_distribution


def get_implicit_combinatorial_adjacency_matrix(stationary_distribution, rw_transition, rw_transition_t):
    number_of_nodes = rw_transition.shape[0]

    sqrtp = stationary_distribution.flatten().real/stationary_distribution.sum()

    sqrtp = sp.sqrt(sqrtp)
    pi_matrix = spsp.spdiags(sqrtp, [0], number_of_nodes, number_of_nodes)

    effective_adjacency_matrix = (pi_matrix.dot(rw_transition) + rw_transition_t.dot(pi_matrix))/2.0

    effective_adjacency_matrix = spsp.coo_matrix(spsp.csr_matrix(effective_adjacency_matrix))
    effective_adjacency_matrix.data = np.real(effective_adjacency_matrix.data)

    effective_adjacency_matrix = spsp.csr_matrix(effective_adjacency_matrix)

    return effective_adjacency_matrix, sqrtp


def get_implicit_directed_adjacency_matrix(stationary_distribution, rw_transition, rho):
    number_of_nodes = rw_transition.shape[0]

    sqrtp = stationary_distribution.flatten().real/stationary_distribution.sum()

    sqrtp = sp.sqrt(sqrtp)
    Q = spsp.spdiags(sqrtp, [0], number_of_nodes, number_of_nodes) * rw_transition * spsp.spdiags(1.0/sqrtp, [0], number_of_nodes, number_of_nodes)

    effective_adjacency_matrix = (1-rho)*(Q + Q.T) /2.0

    effective_adjacency_matrix = spsp.coo_matrix(spsp.csr_matrix(effective_adjacency_matrix))
    effective_adjacency_matrix.data = np.real(effective_adjacency_matrix.data)

    effective_adjacency_matrix = spsp.csr_matrix(effective_adjacency_matrix)

    return effective_adjacency_matrix, np.ones(number_of_nodes, dtype=np.float64)


def get_adjacency_matrix_via_combinatorial_laplacian(adjacency_matrix, rho=0.2):
    number_of_nodes = adjacency_matrix.shape[0]

    ####################################################################################################################
    # Form random walk probability transition matrix
    ####################################################################################################################
    rw_transition, out_degree, in_degree = get_natural_random_walk_matrix(adjacency_matrix, make_shared=False)
    rw_transition = rw_transition.tocoo()
    rw_transition_t = rw_transition.T.tocsr()
    rw_transition = rw_transition.tocsr()

    non_teleportation = 1-rho
    mv = lambda l, v: non_teleportation*l.dot(v) + (rho/number_of_nodes)*np.ones_like(v)
    teleport = lambda vec: mv(rw_transition_t, vec)

    rw_transition_operator = spla.LinearOperator(rw_transition.shape, matvec=teleport, dtype=np.float64)

    ####################################################################################################################
    # Form theta matrix
    ####################################################################################################################
    # Form stationary distribution diagonal matrix
    try:
        eigenvalue, stationary_distribution = spla.eigs(rw_transition_operator,
                                                        k=1,
                                                        which='LM',
                                                        return_eigenvectors=True)
    except spla.ArpackNoConvergence as e:
        print("ARPACK has not converged.")
        eigenvalue = e.eigenvalues
        stationary_distribution = e.eigenvectors
    stationary_distribution = stationary_distribution.flatten().real/stationary_distribution.sum()

    sqrtp = sp.sqrt(stationary_distribution)
    pi_matrix = spsp.spdiags(sqrtp, [0], number_of_nodes, number_of_nodes)

    effective_adjacency_matrix = (pi_matrix.dot(rw_transition) + rw_transition_t.dot(pi_matrix))/2.0

    effective_adjacency_matrix = spsp.coo_matrix(spsp.csr_matrix(effective_adjacency_matrix))
    effective_adjacency_matrix.data = np.real(effective_adjacency_matrix.data)

    return effective_adjacency_matrix, sqrtp


def get_adjacency_matrix_via_directed_laplacian(adjacency_matrix, rho=0.2):
    number_of_nodes = adjacency_matrix.shape[0]

    ####################################################################################################################
    # Form random walk probability transition matrix
    ####################################################################################################################
    rw_transition, out_degree, in_degree = get_natural_random_walk_matrix(adjacency_matrix, make_shared=False)
    rw_transition = rw_transition.tocoo()
    rw_transition_t = rw_transition.T.tocsr()
    rw_transition = rw_transition.tocsr()

    non_teleportation = 1-rho
    mv = lambda l, v: non_teleportation*l.dot(v) + (rho/number_of_nodes)*np.ones_like(v)
    teleport = lambda vec: mv(rw_transition_t, vec)

    rw_transition_operator = spla.LinearOperator(rw_transition.shape, matvec=teleport, dtype=np.float64)

    ####################################################################################################################
    # Form theta matrix
    ####################################################################################################################
    # Form stationary distribution diagonal matrix
    try:
        eigenvalue, stationary_distribution = spla.eigs(rw_transition_operator,
                                                        k=1,
                                                        which='LM',
                                                        return_eigenvectors=True)
    except spla.ArpackNoConvergence as e:
        print("ARPACK has not converged.")
        eigenvalue = e.eigenvalues
        stationary_distribution = e.eigenvectors
    stationary_distribution = stationary_distribution.flatten().real/stationary_distribution.sum()

    sqrtp = sp.sqrt(stationary_distribution)
    Q = spsp.spdiags(sqrtp, [0], number_of_nodes, number_of_nodes) * rw_transition * spsp.spdiags(1.0/sqrtp, [0], number_of_nodes, number_of_nodes)

    effective_adjacency_matrix = (1-rho)*(Q + Q.T) /2.0

    effective_adjacency_matrix = spsp.coo_matrix(spsp.csr_matrix(effective_adjacency_matrix))
    effective_adjacency_matrix.data = np.real(effective_adjacency_matrix.data)

    return effective_adjacency_matrix, np.ones(number_of_nodes, dtype=np.float64)


def zhou_view_weight_calculation(transition_matrix_list, out_degree_list, weights):
    number_of_views = len(transition_matrix_list)
    number_of_nodes = transition_matrix_list[0].shape[0]

    actual_weights = np.empty((number_of_views, number_of_nodes), dtype=np.float64)
    for v in range(number_of_views):
        actual_weights[v, :] = out_degree_list[v]

    actual_weights[actual_weights > 0.0] = 1.0

    for n in range(number_of_nodes):
        row_sum = np.sum(actual_weights[:, n])
        if row_sum > 0.0:
            actual_weights[:, n] = actual_weights[:, n]/row_sum

    for v in range(number_of_views):
        actual_weights[v, :] = actual_weights[v, :]*weights[v]

    return actual_weights


def entropy_view_weight_calculation(adjacency_matrix_list, transition_matrix_list, out_degree_list):
    number_of_views = len(transition_matrix_list)
    number_of_nodes = transition_matrix_list[0].shape[0]

    actual_weights = np.empty((number_of_views, number_of_nodes), dtype=np.float64)
    for v in range(number_of_views):
        actual_weights[v, :] = out_degree_list[v]

    actual_weights[actual_weights > 0.0] = 1.0

    for n in range(number_of_nodes):
        row_nnz_ind = np.where(actual_weights > 0.0)[0]
        if row_nnz_ind.size > 0:
            for v in row_nnz_ind:
                transition_matrix_list


def get_multiview_transition_matrix(adjacency_matrix_list, weights=None, method="zhou"):
    # Get number of matrices.
    number_of_views = len(adjacency_matrix_list)
    if number_of_views < 1:
        print("Empty adjacency matrix list.")
        raise RuntimeError

    # Make sure number of weights is equal to number of matrices.
    if method == "zhou":
        if weights is None:
            weights = (1/number_of_views) * np.ones(number_of_views, dtype=np.float64)
        else:
            if weights.size != number_of_views:
                print("Number of adjacency matrices not equal to number of weights.")
                raise RuntimeError
            else:
                weights = weights/npla.norm(weights, "fro")

    # Make sure all matrices are in csr format.
    adjacency_matrix_list = (spsp.csr_matrix(adjacency_matrix) for adjacency_matrix in adjacency_matrix_list)

    # Get natural random walk transition matrices.
    transition_tuple_list = [get_natural_random_walk_matrix(adjacency_matrix) for adjacency_matrix in adjacency_matrix_list]
    transition_matrix_list = [t[0] for t in transition_tuple_list]
    out_degree_list = [t[1] for t in transition_tuple_list]
    in_degree_list = [t[2] for t in transition_tuple_list]

    # Calculate actual weights for matrices.
    if method == "zhou":
        actual_weights = zhou_view_weight_calculation(transition_matrix_list, out_degree_list, weights)
    elif method == "entropy":
        actual_weights = entropy_view_weight_calculation(adjacency_matrix_list, transition_matrix_list, out_degree_list)
    else:
        print("Invalid view weighting method selected.")
        raise RuntimeError

    # Calculate the multiview implicit transition matrix.
    number_of_nodes = transition_matrix_list[0].shape[0]
    weight_diagonal_matrix = spsp.csr_matrix(spsp.spdiags(actual_weights[0], [0], number_of_nodes, number_of_nodes))
    multiview_implicit_transition_matrix = weight_diagonal_matrix.dot(transition_matrix_list[0])

    for v in range(1, number_of_views):
        weight_diagonal_matrix = spsp.csr_matrix(spsp.spdiags(actual_weights[v], [0], number_of_nodes, number_of_nodes))
        multiview_implicit_transition_matrix += weight_diagonal_matrix.dot(transition_matrix_list[v])

    return multiview_implicit_transition_matrix



__author__ = 'Georgios Rizos (georgerizos@iti.gr)'

import numpy as np
import scipy.sparse as sparse

from reveal_graph_embedding.common import get_file_row_generator


def read_adjacency_matrix(file_path, separator):
    """
    Reads an edge list in csv format and returns the adjacency matrix in SciPy Sparse COOrdinate format.

    Inputs:  - file_path: The path where the adjacency matrix is stored.
             - separator: The delimiter among values (e.g. ",", "\t", " ")

    Outputs: - adjacency_matrix: The adjacency matrix in SciPy Sparse COOrdinate format.
    """
    # Open file
    file_row_generator = get_file_row_generator(file_path, separator)

    # Initialize lists for row and column sparse matrix arguments
    row = list()
    col = list()
    append_row = row.append
    append_col = col.append

    # Read all file rows
    for file_row in file_row_generator:
        source_node = np.int64(file_row[0])
        target_node = np.int64(file_row[1])

        # Add edge
        append_row(source_node)
        append_col(target_node)

        # Since this is an undirected network also add the reciprocal edge
        append_row(target_node)
        append_col(source_node)

    row = np.array(row, dtype=np.int64)
    col = np.array(col, dtype=np.int64)
    data = np.ones_like(row, dtype=np.float64)

    number_of_nodes = np.max(row)  # I assume that there are no missing nodes at the end.

    # Array count should start from 0.
    row -= 1
    col -= 1

    # Form sparse adjacency matrix
    adjacency_matrix = sparse.coo_matrix((data, (row, col)), shape=(number_of_nodes, number_of_nodes))

    return adjacency_matrix


def read_node_label_matrix(file_path, separator, number_of_nodes):
    """
    Reads node-label pairs in csv format and returns a list of tuples and a node-label matrix.

    Inputs:  - file_path: The path where the node-label matrix is stored.
             - separator: The delimiter among values (e.g. ",", "\t", " ")
             - number_of_nodes: The number of nodes of the full graph. It is possible that not all nodes are labelled.

    Outputs: - node_label_matrix: The node-label associations in a NumPy array of tuples format.
             - number_of_categories: The number of categories/classes the nodes may belong to.
             - labelled_node_indices: A NumPy array containing the labelled node indices.
    """
    # Open file
    file_row_generator = get_file_row_generator(file_path, separator)

    # Initialize lists for row and column sparse matrix arguments
    row = list()
    col = list()
    append_row = row.append
    append_col = col.append

    # Populate the arrays
    for file_row in file_row_generator:
        node = np.int64(file_row[0])
        label = np.int64(file_row[1])

        # Add label
        append_row(node)
        append_col(label)

    number_of_categories = len(set(col))  # I assume that there are no missing labels. There may be missing nodes.
    labelled_node_indices = np.array(list(set(row)))

    row = np.array(row, dtype=np.int64)
    col = np.array(col, dtype=np.int64)
    data = np.ones_like(row, dtype=np.float64)

    # Array count should start from 0.
    row -= 1
    col -= 1
    labelled_node_indices -= 1

    # Form sparse adjacency matrix
    node_label_matrix = sparse.coo_matrix((data, (row, col)), shape=(number_of_nodes, number_of_categories))
    node_label_matrix = node_label_matrix.tocsr()

    return node_label_matrix, number_of_categories, labelled_node_indices




__author__ = 'Georgios Rizos (georgerizos@iti.gr)'

try:
    import cPickle as pickle
except ImportError:
    import pickle

import numpy as np
import scipy.sparse as spsp


def get_file_row_generator(file_path, separator, encoding=None):
    """
    Reads an separated value file row by row.

    Inputs: - file_path: The path of the separated value format file.
            - separator: The delimiter among values (e.g. ",", "\t", " ")
            - encoding: The encoding used in the stored text.

    Yields: - words: A list of strings corresponding to each of the file's rows.
    """
    with open(file_path, encoding=encoding) as file_object:
        for line in file_object:
            words = line.strip().split(separator)
            yield words


def store_pickle(file_path, data):
    """
    Pickle some data to a given path.

    Inputs: - file_path: Target file path.
            - data: The python object to be serialized via pickle.
    """
    pkl_file = open(file_path, 'wb')
    pickle.dump(data, pkl_file)
    pkl_file.close()


def load_pickle(file_path):
    """
    Unpickle some data from a given path.

    Input:  - file_path: Target file path.

    Output: - data: The python object that was serialized and stored in disk.
    """
    pkl_file = open(file_path, 'rb')
    data = pickle.load(pkl_file)
    pkl_file.close()
    return data


def read_adjacency_matrix(file_path, separator, undirected):
    """
    Reads an edge list in csv format and returns the adjacency matrix in SciPy Sparse COOrdinate format.

    Inputs:  - file_path: The path where the adjacency matrix is stored.
             - separator: The delimiter among values (e.g. ",", "\t", " ")
             - undirected: If True, create the reciprocal edge for each edge in edge list.

    Outputs: - adjacency_matrix: The adjacency matrix in SciPy Sparse COOrdinate format.
             - node_to_id: A dictionary that maps anonymized node ids to the original node ids.
    """
    # Open file
    file_row_generator = get_file_row_generator(file_path, separator)

    # Initialize lists for row and column sparse matrix arguments.
    row = list()
    col = list()
    data = list()
    append_row = row.append
    append_col = col.append
    append_data = data.append

    # Initialize node anonymizer.
    id_to_node = dict()

    # Read all file rows.
    for file_row in file_row_generator:
        if file_row[0][0] == "#":
            continue

        source_node_id = int(file_row[0])
        target_node_id = int(file_row[1])

        number_of_nodes = len(id_to_node)
        source_node = id_to_node.setdefault(source_node_id,
                                            number_of_nodes)
        number_of_nodes = len(id_to_node)
        target_node = id_to_node.setdefault(target_node_id,
                                            number_of_nodes)

        edge_weight = float(file_row[2])

        # Add edge.
        append_row(source_node)
        append_col(target_node)
        append_data(edge_weight)

        # Since this is an undirected network also add the reciprocal edge.
        if undirected:
            if source_node != target_node:
                append_row(target_node)
                append_col(source_node)
                append_data(edge_weight)

    number_of_nodes = len(id_to_node)
    node_to_id = dict(zip(list(id_to_node.values()),
                          list(id_to_node.keys())))

    row = np.array(row, dtype=np.int64)
    col = np.array(col, dtype=np.int64)
    data = np.array(data, dtype=np.float64)

    # Form sparse adjacency matrix.
    adjacency_matrix = spsp.coo_matrix((data, (row, col)), shape=(number_of_nodes,
                                                                  number_of_nodes))

    return adjacency_matrix, node_to_id


def write_features(file_path,
                   features,
                   separator,
                   node_to_id):
    features = spsp.coo_matrix(features)

    row = features.row
    col = features.col
    data = features.data

    with open(file_path, "w") as f:
        for element in range(row.size):
            node = row[element]
            node_id = node_to_id[node]

            community_id = col[element]

            value = int(data[element])

            file_row = str(node_id) + separator + str(community_id) + separator + str(value) + "\n"
            f.write(file_row)



__author__ = 'Georgios Rizos (georgerizos@iti.gr)'


def pagerank_limit_push(s, r, w_i, a_i, push_node, rho):
    """
    Performs a random step without a self-loop.
    """
    # Calculate the A and B quantities to infinity
    A_inf = rho*r[push_node]
    B_inf = (1-rho)*r[push_node]

    # Update approximate Pagerank and residual vectors
    s[push_node] += A_inf
    r[push_node] = 0.0

    # Update residual vector at push node's adjacent nodes
    r[a_i] += B_inf * w_i


def pagerank_lazy_push(s, r, w_i, a_i, push_node, rho, lazy):
    """
    Performs a random step with a self-loop.

    Introduced in: Andersen, R., Chung, F., & Lang, K. (2006, October).
                   Local graph partitioning using pagerank vectors.
                   In Foundations of Computer Science, 2006. FOCS'06. 47th Annual IEEE Symposium on (pp. 475-486). IEEE.
    """
    # Calculate the A, B and C quantities
    A = rho*r[push_node]
    B = (1-rho)*(1 - lazy)*r[push_node]
    C = (1-rho)*lazy*(r[push_node])

    # Update approximate Pagerank and residual vectors
    s[push_node] += A
    r[push_node] = C

    # Update residual vector at push node's adjacent nodes
    r[a_i] += B * w_i


def cumulative_pagerank_difference_limit_push(s, r, w_i, a_i, push_node, rho):
    """
    Performs a random step without a self-loop.

    Inputs:  - s: A NumPy array that contains the approximate absorbing random walk cumulative probabilities.
             - r: A NumPy array that contains the residual probability distribution.
             - w_i: A NumPy array of probability transition weights from the seed nodes to its adjacent nodes.
             - a_i: A NumPy array of the nodes adjacent to the push node.
             - push_node: The node from which the residual probability is pushed to its adjacent nodes.
             - rho: The restart probability.

    Outputs: - s in 1xn: A NumPy array that contains the approximate absorbing random walk cumulative probabilities.
             - r in 1xn: A NumPy array that contains the residual probability distribution.
    """
    # Calculate the commute quantity
    commute = (1-rho)*r[push_node]

    # Update approximate regularized commute and residual vectors
    r[push_node] = 0.0

    # Update residual vector at push node's adjacent nodes
    commute_probabilities = commute * w_i
    s[a_i] += commute_probabilities
    r[a_i] += commute_probabilities



__author__ = 'Georgios Rizos (georgerizos@iti.gr)'

import numpy as np
import scipy.sparse as sparse
import ctypes as c
import multiprocessing as mp


def get_label_based_random_walk_matrix(adjacency_matrix, labelled_nodes, label_absorption_probability):
    """
    Returns the label-absorbing random walk transition probability matrix.

    Input:  - A: A sparse matrix that contains the adjacency matrix of the graph.

    Output: - W: A sparse matrix that contains the natural random walk transition probability matrix.
    """
    # Turn to sparse.csr_matrix format for faster row access.
    rw_transition = sparse.csr_matrix(adjacency_matrix, dtype=np.float64)

    # Sum along the two axes to get out-degree and in-degree, respectively
    out_degree = rw_transition.sum(axis=1)
    in_degree = rw_transition.sum(axis=0)

    # Form the inverse of the diagonal matrix containing the out-degree
    for i in np.arange(rw_transition.shape[0]):
        rw_transition.data[rw_transition.indptr[i]: rw_transition.indptr[i + 1]] =\
            rw_transition.data[rw_transition.indptr[i]: rw_transition.indptr[i + 1]]/out_degree[i]

    out_degree = np.array(out_degree).astype(np.float64).reshape(out_degree.size)
    in_degree = np.array(in_degree).astype(np.float64).reshape(in_degree.size)

    # When the random walk agent encounters a labelled node, there is a probability that it will be absorbed.
    diag = np.zeros_like(out_degree)
    diag[labelled_nodes] = 1.0
    diag = sparse.dia_matrix((diag, [0]), shape=(in_degree.size, in_degree.size))
    diag = sparse.csr_matrix(diag)

    rw_transition[labelled_nodes, :] = (1-label_absorption_probability)*rw_transition[labelled_nodes, :] + label_absorption_probability*diag[labelled_nodes, :]

    return rw_transition, out_degree, in_degree


def get_natural_random_walk_matrix(adjacency_matrix, make_shared=False):
    """
    Returns the natural random walk transition probability matrix given the adjacency matrix.

    Input:  - A: A sparse matrix that contains the adjacency matrix of the graph.

    Output: - W: A sparse matrix that contains the natural random walk transition probability matrix.
    """
    # Turn to sparse.csr_matrix format for faster row access.
    rw_transition = sparse.csr_matrix(adjacency_matrix, dtype=np.float64, copy=True)

    # Sum along the two axes to get out-degree and in-degree, respectively
    out_degree = rw_transition.sum(axis=1)
    in_degree = rw_transition.sum(axis=0)

    # Form the inverse of the diagonal matrix containing the out-degree
    for i in np.arange(rw_transition.shape[0]):
        rw_transition.data[rw_transition.indptr[i]: rw_transition.indptr[i + 1]] =\
            rw_transition.data[rw_transition.indptr[i]: rw_transition.indptr[i + 1]]/out_degree[i]

    rw_transition.sort_indices()

    out_degree = np.array(out_degree).astype(np.float64).reshape(out_degree.size)
    in_degree = np.array(in_degree).astype(np.float64).reshape(in_degree.size)

    if make_shared:
        number_of_nodes = adjacency_matrix.shape[0]

        out_degree_c = mp.Array(c.c_double, number_of_nodes)
        in_degree_c = mp.Array(c.c_double, number_of_nodes)

        out_degree_shared = np.frombuffer(out_degree_c.get_obj(), dtype=np.float64, count=number_of_nodes)
        in_degree_shared = np.frombuffer(in_degree_c.get_obj(), dtype=np.float64, count=number_of_nodes)

        out_degree_shared[:] = out_degree[:]
        in_degree_shared[:] = in_degree[:]

        indices_c = mp.Array(c.c_int64, rw_transition.indices.size)
        indptr_c = mp.Array(c.c_int64, rw_transition.indptr.size)
        data_c = mp.Array(c.c_double, rw_transition.data.size)

        indices_shared = np.frombuffer(indices_c.get_obj(), dtype=np.int64, count=rw_transition.indices.size)
        indptr_shared = np.frombuffer(indptr_c.get_obj(), dtype=np.int64, count=rw_transition.indptr.size)
        data_shared = np.frombuffer(data_c.get_obj(), dtype=np.float64, count=rw_transition.data.size)

        indices_shared[:] = rw_transition.indices[:]
        indptr_shared[:] = rw_transition.indptr[:]
        data_shared[:] = rw_transition.data[:]

        rw_transition = sparse.csr_matrix((data_shared,
                                           indices_shared,
                                           indptr_shared),
                                          shape=rw_transition.shape)

    return rw_transition, out_degree, in_degree



__author__ = 'Georgios Rizos (georgerizos@iti.gr)'
# cython: profile=True

import numpy as np
cimport numpy as np
cimport cython
import scipy.sparse as sparse

from reveal_graph_embedding.eps_randomwalk.cython_opt.transition import get_natural_random_walk_matrix
from reveal_graph_embedding.eps_randomwalk.cython_opt.similarity import fast_approximate_cumulative_pagerank_difference
from reveal_graph_embedding.embedding.common import normalize_community_features

FLOAT64 = np.float64
ctypedef np.float64_t FLOAT64_t

INT64 = np.int64
ctypedef np.int64_t INT64_t


def arcte(adjacency_matrix, double rho, double epsilon):
    """
    Extracts local community features for all graph nodes based on the partitioning of node-centric similarity vectors.

    Inputs:  - A in R^(nxn): Adjacency matrix of an undirected network represented as a SciPy Sparse COOrdinate matrix.
             - rho: Restart probability
             - epsilon: Approximation threshold

    Outputs: - X in R^(nxC_n): The latent space embedding represented as a SciPy Sparse COOrdinate matrix.
    """
    adjacency_matrix = sparse.csr_matrix(adjacency_matrix)
    number_of_nodes = adjacency_matrix.shape[0]

    # Calculate natural random walk transition probability matrix
    cdef np.ndarray[FLOAT64_t, ndim=1] out_degree
    cdef np.ndarray[FLOAT64_t, ndim=1] in_degree
    rw_transition, out_degree, in_degree = get_natural_random_walk_matrix(adjacency_matrix)

    # Store adjacent nodes and corresponding transition weights in array of arrays form.
    cdef np.ndarray adjacent_nodes = np.ndarray(number_of_nodes, dtype=np.ndarray)
    cdef np.ndarray base_transitions = np.ndarray(number_of_nodes, dtype=np.ndarray)
    for n in np.arange(number_of_nodes):
        rw_transition_row = rw_transition.getrow(n)
        adjacent_nodes[n] = rw_transition_row.indices.astype(np.int64)
        base_transitions[n] = rw_transition_row.data

    # Calculate local communities for all nodes
    row_list = list()
    col_list = list()
    extend_row = row_list.extend
    extend_col = col_list.extend

    cdef long number_of_local_communities = 0

    cdef np.ndarray [FLOAT64_t, ndim=1] s = np.zeros(number_of_nodes, dtype=FLOAT64)
    cdef np.ndarray [FLOAT64_t, ndim=1] r = np.zeros(number_of_nodes, dtype=FLOAT64)

    cdef np.ndarray[INT64_t, ndim=1] iterate_nodes = np.where(out_degree != 0)[0]
    cdef long seed_node
    for seed_node in iterate_nodes:
        print(seed_node)
        s[:] = 0.0
        r[:] = 0.0

        # Calculate similarity matrix slice
        nop = fast_approximate_cumulative_pagerank_difference(s,
                                                              r,
                                                              base_transitions,
                                                              adjacent_nodes,
                                                              out_degree,
                                                              in_degree,
                                                              seed_node,
                                                              rho,
                                                              epsilon)

        s_sparse = sparse.csr_matrix(s, shape=(1, number_of_nodes))

        # Perform degree normalization of approximate similarity matrix slice
        relevant_degrees = in_degree[s_sparse.indices]
        s_sparse.data = np.divide(s_sparse.data, relevant_degrees)

        # Sort the degree normalized approximate similarity matrix slice
        sorted_indices = np.argsort(s_sparse.data, axis=0)
        s_sparse.data = s_sparse.data[sorted_indices]
        s_sparse.indices = s_sparse.indices[sorted_indices]

        # Iterate over the support of the distribution to detect local community
        base_community = set(adjacent_nodes[seed_node])
        base_community.add(seed_node)
        base_community_size = len(base_community)

        base_community_count = 0
        most_unlikely_index = 0
        for i in np.arange(1, s_sparse.data.size + 1):
            if s_sparse.indices[-i] in base_community:
                base_community_count += 1
                if base_community_count == base_community_size:
                    most_unlikely_index = i
                    break

        # Save feature matrix coordinates
        if most_unlikely_index > base_community_count:
            new_rows = s_sparse.indices[-1:-most_unlikely_index-1:-1]
            extend_row(new_rows)
            extend_col(number_of_local_communities*np.ones_like(new_rows))
            number_of_local_communities += 1

    # Form local community feature matrix
    row = np.array(row_list, dtype=np.int64)
    col = np.array(col_list, dtype=np.int64)
    data = np.ones_like(row, dtype=np.float64)
    features = sparse.coo_matrix((data, (row, col)), shape=(number_of_nodes, number_of_local_communities))

    # Form base community feature matrix
    identity_matrix = sparse.csr_matrix(sparse.eye(number_of_nodes, number_of_nodes, dtype=np.float64))
    base_community_features = identity_matrix + adjacency_matrix

    # Stack horizontally matrices to form feature matrix
    features = sparse.hstack([base_community_features, features]).tocoo()

    features = normalize_community_features(features)

    return features


def arcte_and_centrality(adjacency_matrix, double rho, double epsilon):
    """
    Extracts local community features for all graph nodes based on the partitioning of node-centric similarity vectors.

    Inputs:  - A in R^(nxn): Adjacency matrix of an undirected network represented as a SciPy Sparse COOrdinate matrix.
             - rho: Restart probability
             - epsilon: Approximation threshold

    Outputs: - X in R^(nxC_n): The latent space embedding represented as a SciPy Sparse COOrdinate matrix.
             - centrality in R^(nx1): A vector containing the RCT measure of centrality.
    """
    adjacency_matrix = sparse.csr_matrix(adjacency_matrix)
    number_of_nodes = adjacency_matrix.shape[0]

    # Calculate natural random walk transition probability matrix
    cdef np.ndarray[FLOAT64_t, ndim=1] out_degree
    cdef np.ndarray[FLOAT64_t, ndim=1] in_degree
    rw_transition, out_degree, in_degree = get_natural_random_walk_matrix(adjacency_matrix)

    # Store adjacent nodes and corresponding transition weights in array of arrays form.
    cdef np.ndarray adjacent_nodes = np.ndarray(number_of_nodes, dtype=np.ndarray)
    cdef np.ndarray base_transitions = np.ndarray(number_of_nodes, dtype=np.ndarray)
    for n in np.arange(number_of_nodes):
        rw_transition_row = rw_transition.getrow(n)
        adjacent_nodes[n] = rw_transition_row.indices.astype(np.int64)
        base_transitions[n] = rw_transition_row.data

    # Calculate local communities for all nodes
    row_list = list()
    col_list = list()
    extend_row = row_list.extend
    extend_col = col_list.extend

    cdef long number_of_local_communities = 0

    cdef np.ndarray [FLOAT64_t, ndim=1] s = np.zeros(number_of_nodes, dtype=FLOAT64)
    cdef np.ndarray [FLOAT64_t, ndim=1] r = np.zeros(number_of_nodes, dtype=FLOAT64)
    cdef np.ndarray centrality = np.zeros(number_of_nodes, dtype=FLOAT64)

    cdef np.ndarray[INT64_t, ndim=1] iterate_nodes = np.where(out_degree != 0)[0]
    cdef long seed_node
    for seed_node in iterate_nodes:
        # print(seed_node)

        for node in range(number_of_nodes):
            s[node] = 0.0
            r[node] = 0.0

        # Calculate similarity matrix slice
        nop = fast_approximate_cumulative_pagerank_difference(s,
                                                              r,
                                                              base_transitions,
                                                              adjacent_nodes,
                                                              out_degree,
                                                              in_degree,
                                                              seed_node,
                                                              rho,
                                                              epsilon)

        s_sparse = sparse.csr_matrix(s, shape=(1, number_of_nodes))

        # Perform degree normalization of approximate similarity matrix slice
        relevant_degrees = in_degree[s_sparse.indices]
        s_sparse.data = np.divide(s_sparse.data, relevant_degrees)

        # Adjust centrality
        centrality += s_sparse

        # Sort the degree normalized approximate similarity matrix slice
        sorted_indices = np.argsort(s_sparse.data, axis=0)
        s_sparse.data = s_sparse.data[sorted_indices]
        s_sparse.indices = s_sparse.indices[sorted_indices]

        # Iterate over the support of the distribution to detect local community
        base_community = set(adjacent_nodes[seed_node])
        base_community.add(seed_node)
        base_community_size = len(base_community)

        base_community_count = 0
        most_unlikely_index = 0
        for i in np.arange(1, s_sparse.data.size + 1):
            if s_sparse.indices[-i] in base_community:
                base_community_count += 1
                if base_community_count == base_community_size:
                    most_unlikely_index = i
                    break

        # Save feature matrix coordinates
        if most_unlikely_index > base_community_count:
            new_rows = s_sparse.indices[-1:-most_unlikely_index-1:-1]
            extend_row(new_rows)
            extend_col(number_of_local_communities*np.ones_like(new_rows))
            number_of_local_communities += 1

        s_sparse = None

    centrality[np.setdiff1d(np.arange(number_of_nodes), iterate_nodes)] = 1.0

    # Form local community feature matrix
    row = np.array(row_list, dtype=np.int64)
    col = np.array(col_list, dtype=np.int64)
    data = np.ones_like(row, dtype=np.float64)
    features = sparse.coo_matrix((data, (row, col)), shape=(number_of_nodes, number_of_local_communities))

    # Form base community feature matrix
    identity_matrix = sparse.csr_matrix(sparse.eye(number_of_nodes, number_of_nodes, dtype=np.float64))
    base_community_features = identity_matrix + adjacency_matrix

    # Stack horizontally matrices to form feature matrix
    try:
        features = sparse.hstack([base_community_features, features]).tocoo()
    except ValueError as e:
       features = base_community_features

    features = normalize_community_features(features)

    return features, centrality



__author__ = 'Georgios Rizos (georgerizos@iti.gr)'

from sklearn.multiclass import OneVsRestClassifier
from sklearn.svm import LinearSVC
from sklearn.linear_model import LogisticRegression
from sklearn.ensemble import RandomForestClassifier
from sklearn.preprocessing import normalize
import scipy.sparse as spsp
from scipy.sparse import issparse


def model_fit(X_train, y_train, svm_hardness, fit_intercept, number_of_threads, classifier_type="LinearSVC"):
    """
    Fits a Linear Support Vector Classifier to the labelled graph-based features using the LIBLINEAR library.

    One-vs-All: http://scikit-learn.org/stable/modules/generated/sklearn.multiclass.OneVsRestClassifier.html
    LinearSVC:  http://scikit-learn.org/stable/modules/generated/sklearn.svm.LinearSVC.html

    Inputs:  - feature_matrix: The graph based-features in either NumPy or SciPy sparse array format.
             - node_label_matrix: The node-label ground truth in a SciPy sparse matrix format.
             - svm_hardness: Penalty of the error term.
             - fit_intercept: Data centering as per scikit-learn.
             - number_of_threads: The number of threads to use for training the multi-label scheme.
             - classifier_type: A string to be chosen among: * LinearSVC (LibLinear)
                                                             * LogisticRegression (LibLinear)
                                                             * RandomForest

    Output:  - model: A trained scikit-learn One-vs-All multi-label scheme of linear SVC models.
    """
    if classifier_type == "LinearSVC":
        model = OneVsRestClassifier(LinearSVC(C=svm_hardness, random_state=0, dual=False,
                                              fit_intercept=fit_intercept),
                                    n_jobs=number_of_threads)
        model.fit(X_train, y_train)
    elif classifier_type == "LogisticRegression":
        model = OneVsRestClassifier(LogisticRegression(C=svm_hardness, random_state=0, dual=False,
                                                       fit_intercept=fit_intercept),
                                    n_jobs=number_of_threads)
        model.fit(X_train, y_train)
    elif classifier_type == "RandomForest":
        model = OneVsRestClassifier(RandomForestClassifier(n_estimators=1000, criterion="gini",
                                       n_jobs=number_of_threads, random_state=0))
        if issparse(X_train):
            model.fit(X_train.tocsc(), y_train.toarray())
        else:
            model.fit(X_train, y_train.toarray())
    else:
        print("Invalid classifier type.")
        raise RuntimeError

    return model


def weigh_users(X_test, model, classifier_type="LinearSVC"):
    """
    Uses a trained model and the unlabelled features to produce a user-to-label distance matrix.

    Inputs:  - feature_matrix: The graph based-features in either NumPy or SciPy sparse array format.
             - model: A trained scikit-learn One-vs-All multi-label scheme of linear SVC models.
             - classifier_type: A string to be chosen among: * LinearSVC (LibLinear)
                                                             * LogisticRegression (LibLinear)
                                                             * RandomForest

    Output:  - decision_weights: A NumPy array containing the distance of each user from each label discriminator.
    """
    if classifier_type == "LinearSVC":
        decision_weights = model.decision_function(X_test)
    elif classifier_type == "LogisticRegression":
        decision_weights = model.decision_function(X_test)
    elif classifier_type == "RandomForest":
        # if issparse(X_test):
        #     decision_weights = np.hstack(a[:, 1].reshape(X_test.shape[0], 1) for a in model.predict_proba(X_test.tocsr()))
        # else:
        #     decision_weights = np.hstack(a[:, 1].reshape(X_test.shape[0], 1) for a in model.predict_proba(X_test))
        if issparse(X_test):
            decision_weights = model.predict_proba(X_test.tocsr())
        else:
            decision_weights = model.predict_proba(X_test)
    else:
        print("Invalid classifier type.")
        raise RuntimeError

    return decision_weights


def classify_users(X_test, model, classifier_type="LinearSVC"):
    """
    Uses a trained model and the unlabelled features to associate users with labels.

    The decision is done as per scikit-learn:
    http://scikit-learn.org/stable/modules/generated/sklearn.multiclass.OneVsRestClassifier.html
    http://scikit-learn.org/stable/modules/generated/sklearn.svm.LinearSVC.html#sklearn.svm.LinearSVC.predict

    Inputs:  - feature_matrix: The graph based-features in either NumPy or SciPy sparse array format.
             - model: A trained scikit-learn One-vs-All multi-label scheme of linear SVC models.

    Output:  - decision_weights: A NumPy array containing the distance of each user from each label discriminator.
    """
    if classifier_type == "LinearSVC":
        prediction = model.decision_function(X_test)

        prediction[prediction <= 0.0] = 0.0
        prediction = spsp.coo_matrix(prediction)

        prediction = normalize(prediction, norm="l2", axis=0)
    elif classifier_type == "LogisticRegression":
        prediction = model.predict_proba(X_test)

        prediction[prediction <= 0.4] = 0.0
        prediction = spsp.coo_matrix(prediction)
    elif classifier_type == "RandomForest":
        if issparse(X_test):
            prediction = model.predict_proba(X_test.tocsr())
        else:
            prediction = model.predict_proba(X_test)

        prediction[prediction <= 0.4] = 0.0
        prediction = spsp.coo_matrix(prediction)
    else:
        print("Invalid classifier type.")
        raise RuntimeError

    return prediction



__author__ = 'Georgios Rizos (georgerizos@iti.gr)'

import numpy as np


def calculate_entropy(p):
    """
    Calculates entropy in bits of a probability distribution.

    Input:  - p: A probability distribution in numpy array format.

    Output: - entropy: The statistical entropy of distribution p.
    """
    entropy = - np.sum(np.multiply(p, np.log2(p)))

    return entropy


def calculate_cross_entropy(p, q):
    """
    Calculates cross-entropy in bits of an approximate probability distribution q from p.

    Input:  - p: The source probability distribution in numpy array format.
            - q: An approximate probability distribution in numpy array format.

    Output: - cross_entropy: The cross-entropy of distribution q from p.
    """
    cross_entropy = - np.sum(np.multiply(p, np.log2(q)))

    return cross_entropy


def calculate_kl_divergence(p, q):
    """
    Calculates KL-divergence in bits of an approximate probability distribution q from p.

    Input:  - p: The source probability distribution in numpy array format.
            - q: An approximate probability distribution in numpy array format.

    Output: - p_q_kl_divergence: The KL-divergence of distribution q from p.
    """
    # Normalize p distribution.
    p /= p.sum()

    p_entropy = calculate_entropy(p)

    # Normalize q distribution.
    q /= q.sum()

    p_q_cross_entropy = calculate_cross_entropy(p, q)

    p_q_kl_divergence = p_q_cross_entropy - p_entropy

    return p_q_kl_divergence


def calculate_kl_divergence_stream(p, q_gen):
    # Normalize p distribution.
    p /= p.sum()

    p_entropy = calculate_entropy(p)

    for q in q_gen:
        # Normalize q distribution.
        q /= q.sum()

        p_q_cross_entropy = calculate_cross_entropy(p, q)

        p_q_kl_divergence = p_q_cross_entropy - p_entropy

        yield p_q_kl_divergence






__author__ = 'Georgios Rizos (georgerizos@iti.gr)'

import os
import errno


def make_sure_path_exists(path):
    """
    Checks if a directory path exists, otherwise it makes it.

    Input: - path: A string containing a directory path.
    """
    try:
        os.makedirs(path)
    except OSError as exception:
        if exception.errno != errno.EEXIST:
            raise



__author__ = 'Georgios Rizos (georgerizos@iti.gr)'

import numpy as np
import scipy.sparse as sparse
from sklearn.metrics import precision_recall_fscore_support
from sklearn.metrics import f1_score


def form_node_label_prediction_matrix(y_pred, y_test):
    """
    Given the discriminator distances, this function forms the node-label prediction matrix.

    It is assumed that the number of true labels is known.

    Inputs:  - y_pred: A NumPy array that contains the distance from the discriminator for each label for each user.
             - y_test: The node-label ground truth for the test set in a SciPy sparse CSR matrix format.

    Outputs: - y_pred: The node-label prediction for the test set in a SciPy sparse CSR matrix format.
    """
    number_of_test_nodes = y_pred.shape[0]

    # We calculate the number of true labels for each node.
    true_number_of_labels = np.squeeze(y_test.sum(axis=1))

    # We sort the prediction array for each node.
    index = np.argsort(y_pred, axis=1)

    row = np.empty(y_test.getnnz(), dtype=np.int64)
    col = np.empty(y_test.getnnz(), dtype=np.int64)
    start = 0
    for n in np.arange(number_of_test_nodes):
        end = start + true_number_of_labels[0, n]
        row[start:end] = n
        col[start:end] = index[n, -1:-true_number_of_labels[0, n]-1:-1]
        start = end
    data = np.ones_like(row, dtype=np.int8)

    y_pred = sparse.coo_matrix((data, (row, col)), shape=y_test.shape)

    return y_pred


def calculate_measures(y_pred, y_test):
    """
    Calculates the F-scores and F-score averages given a classification result and a ground truth.

    Inputs:  - y_pred: The node-label prediction for the test set in a SciPy sparse CSR matrix format.
             - y_test: The node-label ground truth for the test set in a SciPy sparse CSR matrix format.

    Outputs: - measures: A number of NumPy arrays containing evaluation scores for the experiment.
    """
    y_pred = y_pred.toarray()
    y_test = y_test.toarray()

    macro_precision, macro_recall, macro_F1, macro_support = precision_recall_fscore_support(y_test,
                                                                                             y_pred,
                                                                                             beta=1.0,
                                                                                             average="macro")

    micro_precision, micro_recall, micro_F1, micro_support = precision_recall_fscore_support(y_test,
                                                                                             y_pred,
                                                                                             beta=1.0,
                                                                                             average="micro")

    F1 = f1_score(y_test,
                  y_pred,
                  average=None)

    measures = [macro_recall, micro_recall, macro_precision, micro_precision, macro_F1, micro_F1, F1]
    return measures



__author__ = 'Georgios Rizos (georgerizos@iti.gr)'

import numpy as np
import scipy.sparse as spsp

from reveal_graph_embedding.embedding.implicit import get_adjacency_matrix_via_combinatorial_laplacian,\
    get_adjacency_matrix_via_directed_laplacian


def get_unnormalized_laplacian(adjacency_matrix):
    # Calculate diagonal matrix of node degrees.
    degree = spsp.dia_matrix((adjacency_matrix.sum(axis=0), np.array([0])), shape=adjacency_matrix.shape)
    degree = degree.tocsr()

    # Calculate sparse graph Laplacian.
    laplacian = spsp.csr_matrix(-adjacency_matrix + degree, dtype=np.float64)

    return laplacian


def get_normalized_laplacian(adjacency_matrix):
    # Calculate diagonal matrix of node degrees.
    degree = spsp.dia_matrix((adjacency_matrix.sum(axis=0), np.array([0])), shape=adjacency_matrix.shape)
    degree = degree.tocsr()

    # Calculate sparse graph Laplacian.
    adjacency_matrix = spsp.csr_matrix(-adjacency_matrix + degree, dtype=np.float64)

    # Calculate inverse square root of diagonal matrix of node degrees.
    degree.data = np.real(1/np.sqrt(degree.data))

    # Calculate sparse normalized graph Laplacian.
    normalized_laplacian = degree*adjacency_matrix*degree

    return normalized_laplacian


def get_random_walk_laplacian(adjacency_matrix):
    # Calculate diagonal matrix of node degrees.
    degree = spsp.dia_matrix((adjacency_matrix.sum(axis=0), np.array([0])), shape=adjacency_matrix.shape)
    degree = degree.tocsr()

    # Calculate sparse graph Laplacian.
    adjacency_matrix = spsp.csr_matrix(-adjacency_matrix + degree, dtype=np.float64)

    # Calculate inverse of diagonal matrix of node degrees.
    degree.data = np.real(1/degree.data)

    # Calculate sparse normalized graph Laplacian.
    random_walk_laplacian = degree*adjacency_matrix

    return random_walk_laplacian


def get_directed_laplacian(adjacency_matrix, rho=0.2):
    number_of_nodes = adjacency_matrix.shape[0]

    effective_adjacency_matrix, rw_distribution = get_adjacency_matrix_via_directed_laplacian(adjacency_matrix, rho)

    I = spsp.spdiags(rw_distribution, [0], number_of_nodes, number_of_nodes)
    theta_matrix = I - effective_adjacency_matrix

    return theta_matrix


def get_combinatorial_laplacian(adjacency_matrix, rho=0.2):
    number_of_nodes = adjacency_matrix.shape[0]

    effective_adjacency_matrix, rw_distribution = get_adjacency_matrix_via_combinatorial_laplacian(adjacency_matrix, rho)

    I = spsp.spdiags(rw_distribution, [0], number_of_nodes, number_of_nodes)
    theta_matrix = I - effective_adjacency_matrix

    return theta_matrix



__author__ = 'Georgios Rizos (georgerizos@iti.gr)'

import os
import numpy as np

from reveal_graph_embedding.common import get_file_row_generator


def write_results(performance_measures, target_file_path):
    with open(target_file_path, "w") as fp:
        first_row = "*** Percentages:" + "\n"
        fp.write(first_row)

        second_row = "1\t2\t3\t4\t5\t6\t7\t8\t9\t10" + "\n"
        fp.write(second_row)

        fp.write("\n\n")
        write_average_score_row(fp, "Macro F1", performance_measures[4])

        fp.write("\n\n")
        write_average_score_row(fp, "Micro F1", performance_measures[5])


def store_performace_measures(performance_measures, memory_path, experiment_string):
    # Unpack performance measures
    # mean_macro_precision = performance_measures[0][0]
    # std_macro_precision = performance_measures[0][1]
    # mean_micro_precision = performance_measures[1][0]
    # std_micro_precision = performance_measures[1][1]
    # mean_macro_recall = performance_measures[2][0]
    # std_macro_recall = performance_measures[2][1]
    # mean_micro_recall = performance_measures[3][0]
    # std_micro_recall = performance_measures[3][1]
    # mean_macro_F1 = performance_measures[4][0]
    # std_macro_F1 = performance_measures[4][1]
    # mean_micro_F1 = performance_measures[5][0]
    # std_micro_F1 = performance_measures[5][1]
    F1 = performance_measures[6]

    number_of_categories = F1.shape[1]

    # Store average scores
    path = memory_path + "/scores/" + experiment_string + "_average_scores.txt"
    if not os.path.exists(path):
        with open(path, "w") as fp:
            write_average_score_row(fp, "Macro Precision", performance_measures[0])
            fp.write("\n\n")

            write_average_score_row(fp, "Micro Precision", performance_measures[1])
            fp.write("\n\n")

            write_average_score_row(fp, "Macro Recall", performance_measures[2])
            fp.write("\n\n")

            write_average_score_row(fp, "Micro Recall", performance_measures[3])
            fp.write("\n\n")

            write_average_score_row(fp, "Macro F1", performance_measures[4])
            fp.write("\n\n")

            write_average_score_row(fp, "Micro F1", performance_measures[5])

    # Store category-specific F scores
    path = memory_path + "/scores/" + experiment_string + "_F_scores.txt"
    if not os.path.exists(path):
        with open(path, "w") as fp:
            for c in np.arange(number_of_categories):
                row = list(F1[:, c])
                row = [str(score) for score in row]
                row = "\t".join(row) + "\n"
                fp.write(row)


def write_average_score_row(fp, score_name, scores):
    """
    Simple utility function that writes an average score row in a file designated by a file pointer.

    Inputs:  - fp: A file pointer.
             - score_name: What it says on the tin.
             - scores: An array of average score values corresponding to each of the training set percentages.
    """
    row = "--" + score_name + "--"
    fp.write(row)
    for vector in scores:
        row = list(vector)
        row = [str(score) for score in row]
        row = "\n" + "\t".join(row)
        fp.write(row)


def read_performance_measures(file_path, number=10):
    file_row_gen = get_file_row_generator(file_path, "\t")

    F1_macro_mean = np.zeros(number, dtype=np.float64)
    F1_macro_std = np.zeros(number, dtype=np.float64)
    F1_micro_mean = np.zeros(number, dtype=np.float64)
    F1_micro_std = np.zeros(number, dtype=np.float64)

    for r in range(18):
        file_row = next(file_row_gen)

    file_row = [float(score) for score in file_row]
    F1_macro_mean[:] = file_row

    file_row = next(file_row_gen)
    file_row = [float(score) for score in file_row]
    F1_macro_std[:] = file_row

    for r in range(3):
        file_row = next(file_row_gen)

    file_row = [float(score) for score in file_row]
    F1_micro_mean[:] = file_row

    file_row = next(file_row_gen)
    file_row = [float(score) for score in file_row]
    F1_micro_std[:] = file_row

    return F1_macro_mean, F1_macro_std, F1_micro_mean, F1_micro_std



__author__ = 'Georgios Rizos (georgerizos@iti.gr)'

from collections import deque
import numpy as np

from reveal_graph_embedding.eps_randomwalk.push import pagerank_limit_push
from reveal_graph_embedding.eps_randomwalk.push import pagerank_lazy_push
from reveal_graph_embedding.eps_randomwalk.push import cumulative_pagerank_difference_limit_push


def fast_approximate_personalized_pagerank(s,
                                           r,
                                           w_i,
                                           a_i,
                                           out_degree,
                                           in_degree,
                                           seed_node,
                                           rho=0.2,
                                           epsilon=0.00001):
    """
    Calculates the approximate personalized PageRank starting from a seed node without self-loops.
    """
    # Initialize approximate PageRank and residual distributions
    # s = np.zeros(number_of_nodes, dtype=np.float64)
    # r = np.zeros(number_of_nodes, dtype=np.float64)
    r[seed_node] = 1.0

    # Initialize queue of nodes to be pushed
    pushable = deque()
    pushable.append(seed_node)

    # Do one push anyway
    push_node = pushable.popleft()

    pagerank_limit_push(s,
                        r,
                        w_i[push_node],
                        a_i[push_node],
                        push_node,
                        rho)
    number_of_push_operations = 1

    i = np.where(np.divide(r[a_i[push_node]], in_degree[a_i[push_node]]) >= epsilon)[0]
    if i.size > 0:
        pushable.extend(a_i[push_node][i])

    while len(pushable) > 0:
        # While there are nodes with large residual probabilities, push
        push_node = pushable.popleft()
        if r[push_node]/in_degree[push_node] >= epsilon:
            pagerank_limit_push(s,
                                r,
                                w_i[push_node],
                                a_i[push_node],
                                push_node,
                                rho)
            number_of_push_operations += 1

            i = np.where(np.divide(r[a_i[push_node]], in_degree[a_i[push_node]]) >= epsilon)[0]
            if i.size > 0:
                pushable.extend(a_i[push_node][i])

    return number_of_push_operations


def lazy_approximate_personalized_pagerank(s,
                                           r,
                                           w_i,
                                           a_i,
                                           out_degree,
                                           in_degree,
                                           seed_node,
                                           rho=0.2,
                                           epsilon=0.00001,
                                           laziness_factor=0.5):
    """
    Calculates the approximate personalized PageRank starting from a seed node with self-loops.

    Introduced in: Andersen, R., Chung, F., & Lang, K. (2006, October).
                   Local graph partitioning using pagerank vectors.
                   In Foundations of Computer Science, 2006. FOCS'06. 47th Annual IEEE Symposium on (pp. 475-486). IEEE.
    """
    # Initialize approximate PageRank and residual distributions
    # s = np.zeros(number_of_nodes, dtype=np.float64)
    # r = np.zeros(number_of_nodes, dtype=np.float64)
    r[seed_node] = 1.0

    # Initialize queue of nodes to be pushed
    pushable = deque()
    pushable.append(seed_node)

    # Do one push anyway
    push_node = pushable.popleft()

    pagerank_lazy_push(s,
                       r,
                       w_i[push_node],
                       a_i[push_node],
                       push_node,
                       rho,
                       laziness_factor)
    number_of_push_operations = 1

    i = np.where(np.divide(r[a_i[push_node]], in_degree[a_i[push_node]]) >= epsilon)[0]
    if i.size > 0:
        pushable.extend(a_i[push_node][i])

    while r[push_node]/in_degree[push_node] >= epsilon:
        pagerank_lazy_push(s,
                           r,
                           w_i[push_node],
                           a_i[push_node],
                           push_node,
                           rho,
                           laziness_factor)
        number_of_push_operations += 1

    # While there are nodes with large residual probabilities, push
    while len(pushable) > 0:
        push_node = pushable.popleft()

        if r[push_node]/in_degree[push_node] >= epsilon:
            pagerank_lazy_push(s,
                               r,
                               w_i[push_node],
                               a_i[push_node],
                               push_node,
                               rho,
                               laziness_factor)
            number_of_push_operations += 1

            i = np.where(np.divide(r[a_i[push_node]], in_degree[a_i[push_node]]) >= epsilon)[0]
            if i.size > 0:
                pushable.extend(a_i[push_node][i])

        while r[push_node]/in_degree[push_node] >= epsilon:
            pagerank_lazy_push(s,
                               r,
                               w_i[push_node],
                               a_i[push_node],
                               push_node,
                               rho,
                               laziness_factor)
            number_of_push_operations += 1

    return number_of_push_operations


def fast_approximate_cumulative_pagerank_difference(s,
                                                    r,
                                                    w_i,
                                                    a_i,
                                                    out_degree,
                                                    in_degree,
                                                    seed_node,
                                                    rho=0.2,
                                                    epsilon=0.00001):
    """
    Calculates cumulative PageRank difference probability starting from a seed node without self-loops.

    Inputs:  - w_i: A NumPy array of arrays of probability transition weights from the seed nodes to its adjacent nodes.
             - a_i: A NumPy array of arrays of the nodes adjacent to the seed node.
             - out_degree: A NumPy array of node out_degrees.
             - in_degree: A NumPy array of node in_degrees.
             - seed_node: The seed for the node-centric personalized PageRank.
             - rho: The restart probability. Usually set in [0.1, 0.2].
             - epsilon: The error threshold.

    Outputs: - s in 1xn: A sparse vector that contains the approximate absorbing random walk cumulative probabilities.
             - r in 1xn: A sparse vector that contains the residual probability distribution.
             - nop: The number of limit probability push operations performed.
    """
    # Initialize the similarity matrix slice and the residual distribution
    # s = np.zeros(number_of_nodes, dtype=np.float64)
    # r = np.zeros(number_of_nodes, dtype=np.float64)
    s[seed_node] = 1.0
    r[seed_node] = 1.0

    # Initialize double-ended queue of nodes to be pushed
    pushable = deque()
    pushable.append(seed_node)

    # Do one push for free
    push_node = pushable.popleft()

    cumulative_pagerank_difference_limit_push(s,
                                              r,
                                              w_i[push_node],
                                              a_i[push_node],
                                              push_node,
                                              rho)
    number_of_push_operations = 1

    i = np.where(np.divide(r[a_i[push_node]], in_degree[a_i[push_node]]) >= epsilon)[0]
    if i.size > 0:
        pushable.extend(a_i[push_node][i])

    # While there are nodes with large residual probabilities, push
    while len(pushable) > 0:
        push_node = pushable.popleft()

        # If the threshold is not satisfied, perform a push operation
        # Both this and the later check are needed, since the pushable queue may contain duplicates.
        if r[push_node]/in_degree[push_node] >= epsilon:
            cumulative_pagerank_difference_limit_push(s,
                                                      r,
                                                      w_i[push_node],
                                                      a_i[push_node],
                                                      push_node,
                                                      rho)
            number_of_push_operations += 1

            # Update pushable double-ended queue
            i = np.where(np.divide(r[a_i[push_node]], in_degree[a_i[push_node]]) >= epsilon)[0]
            if i.size > 0:
                pushable.extend(a_i[push_node][i])

    # Sparsify and return.
    # s_sparse = sparse.csr_matrix(s, shape=(1, number_of_nodes))
    # r_sparse = sparse.csr_matrix(r, shape=(1, number_of_nodes))

    return number_of_push_operations



__author__ = 'Georgios Rizos (georgerizos@iti.gr)'

import cProfile as profile
import pstats
import gc
import os
import operator
import numpy as np
from reveal_user_classification.eps_randomwalk import similarity as similarity
from reveal_user_classification.eps_randomwalk.transition import get_natural_random_walk_matrix


# from reveal_user_classification.eps_randomwalk.cython_opt import similarity as csimilarity
# from reveal_user_classification.eps_randomwalk.cython_opt.similarity import *
# from reveal_user_classification.eps_randomwalk.cython_opt.push import *
from reveal_user_classification.embedding.arcte.arcte import calculate_epsilon_effective


def similarity_slice_benchmark(number_of_trials,
                               adjacency_matrix,
                               rho_effective,
                               epsilon,
                               laziness_factor,
                               profile_individual_stat_folder):
    """
    Compares the efficiency of approaches calculating similarity matrix slices.
    """
    adjacency_matrix = adjacency_matrix.tocsr()
    number_of_nodes = adjacency_matrix.shape[0]

    # Calculate random walk transition probability matrix.
    rw_transition, out_degree, in_degree = get_natural_random_walk_matrix(adjacency_matrix)

    # Calculate base communities(ego excluded) and out-degrees.
    adjacent_nodes = np.ndarray(number_of_nodes, dtype=np.ndarray)
    base_transitions = np.ndarray(number_of_nodes, dtype=np.ndarray)
    for n in np.arange(number_of_nodes):
        rw_transition_row = rw_transition.getrow(n)
        adjacent_nodes[n] = rw_transition_row.indices.astype(np.int64)
        base_transitions[n] = rw_transition_row.data

    # Calculate restart probability in the case of lazy PageRank.
    rho = (rho_effective * (1 - laziness_factor))/(1 - laziness_factor * rho_effective)

    s = np.zeros(number_of_nodes, dtype=np.float64)
    r = np.zeros(number_of_nodes, dtype=np.float64)

    iterate_nodes = np.where(out_degree != 0)[0]

    for trial in range(number_of_trials):
        gc.collect()
        print("Trial #", trial+1)
        for n, node in enumerate(iterate_nodes):

            if n % 500 == 0:
                gc.collect()

            epsilon_effective = calculate_epsilon_effective(None,
                                                            epsilon,
                                                            out_degree[node],
                                                            out_degree[adjacent_nodes[node]],
                                                            None)

            # Perform fast PageRank.
            s[:] = 0.0
            r[:] = 0.0
            filename = profile_individual_stat_folder + "/fpr_vertex_" + str(node) + "_trial_" + str(trial) + ".stats"
            profile.runctx("""similarity.fast_approximate_personalized_pagerank(s,
                                                                                r,
                                                                                base_transitions,
                                                                                adjacent_nodes,
                                                                                out_degree,
                                                                                in_degree,
                                                                                node,
                                                                                rho_effective,
                                                                                epsilon)""",
                           globals(),
                           {"s": s,
                            "r": r,
                            "base_transitions": base_transitions,
                            "adjacent_nodes": adjacent_nodes,
                            "out_degree": out_degree,
                            "in_degree": in_degree,
                            "node": node,
                            "rho_effective": rho_effective,
                            "epsilon": epsilon_effective},
                           filename=filename)

            # Perform lazy PageRank
            s[:] = 0.0
            r[:] = 0.0
            filename = profile_individual_stat_folder + "/lpr_vertex_" + str(node) + "_trial_" + str(trial) + ".stats"
            profile.runctx("""similarity.lazy_approximate_personalized_pagerank(s,
                                                                                r,
                                                                                base_transitions,
                                                                                adjacent_nodes,
                                                                                out_degree,
                                                                                in_degree,
                                                                                node,
                                                                                rho,
                                                                                epsilon)""",
                           globals(),
                           {"s": s,
                            "r": r,
                            "base_transitions": base_transitions,
                            "adjacent_nodes": adjacent_nodes,
                            "out_degree": out_degree,
                            "in_degree": in_degree,
                            "node": node,
                            "rho": rho,
                            "epsilon": epsilon_effective},
                           filename=filename)

            # Perform Regularized Commute-Time
            s[:] = 0.0
            r[:] = 0.0
            filename = profile_individual_stat_folder + "/fcprd_vertex_" + str(node) + "_trial_" + str(trial) + ".stats"
            profile.runctx("""similarity.fast_approximate_regularized_commute(s,
                                                                              r,
                                                                              base_transitions,
                                                                              adjacent_nodes,
                                                                              out_degree,
                                                                              in_degree,
                                                                              node,
                                                                              rho_effective,
                                                                              epsilon)""",
                           globals(),
                           {"s": s,
                            "r": r,
                            "base_transitions": base_transitions,
                            "adjacent_nodes": adjacent_nodes,
                            "out_degree": out_degree,
                            "in_degree": in_degree,
                            "node": node,
                            "rho_effective": rho_effective,
                            "epsilon": epsilon_effective},
                           filename=filename)


def aggregate_measurements(number_of_trials,
                           adjacency_matrix,
                           profile_individual_stat_folder,
                           profile_aggregate_stat_folder):
    adjacency_matrix = adjacency_matrix.tocsr()
    number_of_nodes = adjacency_matrix.shape[0]

    # Calculate random walk transition probability matrix.
    rw_transition, out_degree, in_degree = get_natural_random_walk_matrix(adjacency_matrix)

    iterate_nodes = np.where(out_degree != 0)[0]

    fpr_stats = pstats.Stats()
    lpr_stats = pstats.Stats()
    fcprd_stats = pstats.Stats()

    for trial in range(number_of_trials):
        for n in range(iterate_nodes.size):
            node = iterate_nodes[n]

            # Fast PR filename.
            filename = profile_individual_stat_folder + "/fpr_vertex_" + str(node) + "_trial_" + str(trial) + ".stats"
            fpr_stats.add(filename)
            # os.remove(filename)

            # Lazy PR filename.
            filename = profile_individual_stat_folder + "/lpr_vertex_" + str(node) + "_trial_" + str(trial) + ".stats"
            lpr_stats.add(filename)
            # os.remove(filename)

            # Fast CPRD filename.
            filename = profile_individual_stat_folder + "/fcprd_vertex_" + str(node) + "_trial_" + str(trial) + ".stats"
            fcprd_stats.add(filename)
            # os.remove(filename)

    fpr_stats.dump_stats(profile_aggregate_stat_folder + "/fpr_vertex.stats")
    lpr_stats.dump_stats(profile_aggregate_stat_folder + "/lpr_vertex.stats")
    fcprd_stats.dump_stats(profile_aggregate_stat_folder + "/fcprd_vertex.stats")


def print_stats(number_of_trials,
                profile_aggregate_stat_folder,
                profile_output_stat_folder):
    lpr_stats = pstats.Stats(profile_aggregate_stat_folder + "/lpr_vertex.stats")
    lpr_output_stats = extract_output_stats(number_of_trials, lpr_stats, ["pagerank_lazy_push", "lazy_approximate_personalized_pagerank"], None)

    fpr_stats = pstats.Stats(profile_aggregate_stat_folder + "/fpr_vertex.stats")
    fcprd_stats = pstats.Stats(profile_aggregate_stat_folder + "/fcprd_vertex.stats")

    fpr_output_stats = extract_output_stats(number_of_trials, fpr_stats, ["pagerank_limit_push", "fast_approximate_personalized_pagerank"], lpr_output_stats)
    fcprd_output_stats = extract_output_stats(number_of_trials, fcprd_stats, ["regularized_limit_commute", "fast_approximate_regularized_commute"], lpr_output_stats)

    target_file = profile_output_stat_folder + "/output_stats.txt"

    write_output_stats(target_file, lpr_output_stats, fpr_output_stats, fcprd_output_stats)


def extract_output_stats(number_of_trials, aggregate_stats_object, function_list, base_stats):
    output_stats = dict()

    aggregate_stats_object.strip_dirs().sort_stats(-1).print_stats()
    s = aggregate_stats_object.stats

    push_key = [tup for tup in s.keys() if operator.getitem(tup, 2) == function_list[0]][0]
    similarity_key = [tup for tup in s.keys() if operator.getitem(tup, 2) == function_list[1]][0]

    push_value = s[push_key]
    similarity_value = s[similarity_key]

    number_of_nodes = similarity_value[1]/number_of_trials

    output_stats["total_number_of_operations"] = push_value[1]/number_of_trials
    output_stats["total_time"] = push_value[3]/number_of_trials
    output_stats["per_node_number_of_operations"] = output_stats["total_number_of_operations"]/number_of_nodes
    output_stats["per_node_time"] = output_stats["total_time"]/number_of_nodes

    if base_stats is not None:
        output_stats["speedup_number_of_operations"] = percentage_increase(output_stats["total_number_of_operations"],
                                                                           base_stats["total_number_of_operations"])
        output_stats["speedup_time"] = percentage_increase(output_stats["total_time"],
                                                           base_stats["total_time"])

    print(output_stats)

    return output_stats


def percentage_increase(fast, lazy):
    return (np.abs(fast - lazy)/lazy)*100


def write_output_stats(target_file, lpr_output_stats, fpr_output_stats, fcprd_output_stats):
    with open(target_file, "w") as fp:
        row = "*** Lazy PageRank ***" + "\n"
        fp.write(row)

        row = "total nop:" + "\t" + str(lpr_output_stats["total_number_of_operations"]) + "\n"
        fp.write(row)
        row = "total time:" + "\t" + str(lpr_output_stats["total_time"]) + "\n"
        fp.write(row)
        row = "per node nop:" + "\t" + str(lpr_output_stats["per_node_number_of_operations"]) + "\n"
        fp.write(row)
        row = "per node time:" + "\t" + str(lpr_output_stats["per_node_time"]) + "\n"
        fp.write(row)

        row = "\n"
        fp.write(row)

        row = "*** Fast PageRank ***" + "\n"
        fp.write(row)

        row = "total nop:" + "\t" + str(fpr_output_stats["total_number_of_operations"]) + "\n"
        fp.write(row)
        row = "total time:" + "\t" + str(fpr_output_stats["total_time"]) + "\n"
        fp.write(row)
        row = "per node nop:" + "\t" + str(fpr_output_stats["per_node_number_of_operations"]) + "\n"
        fp.write(row)
        row = "per node time:" + "\t" + str(fpr_output_stats["per_node_time"]) + "\n"
        fp.write(row)
        row = "speedup nop:" + "\t" + str(fpr_output_stats["speedup_number_of_operations"]) + "\n"
        fp.write(row)
        row = "speedup time:" + "\t" + str(fpr_output_stats["speedup_time"]) + "\n"
        fp.write(row)

        row = "\n"
        fp.write(row)

        row = "*** Fast Cumulative PageRank Differences ***" + "\n"
        fp.write(row)

        row = "total nop:" + "\t" + str(fcprd_output_stats["total_number_of_operations"]) + "\n"
        fp.write(row)
        row = "total time:" + "\t" + str(fcprd_output_stats["total_time"]) + "\n"
        fp.write(row)
        row = "per node nop:" + "\t" + str(fcprd_output_stats["per_node_number_of_operations"]) + "\n"
        fp.write(row)
        row = "per node time:" + "\t" + str(fcprd_output_stats["per_node_time"]) + "\n"
        fp.write(row)
        row = "speedup nop:" + "\t" + str(fcprd_output_stats["speedup_number_of_operations"]) + "\n"
        fp.write(row)
        row = "speedup time:" + "\t" + str(fcprd_output_stats["speedup_time"]) + "\n"
        fp.write(row)

        row = "\n"
        fp.write(row)


# def similarity_slice_benchmark(adjacency_matrix, alpha_effective, epsilon, laziness_factor):
#     """
#     Compares the efficiency of approaches calculating similarity matrix slices.
#     """
#     adjacency_matrix = adjacency_matrix.tocsr()
#     number_of_nodes = adjacency_matrix.shape[0]
#
#     # Calculate random walk transition probability matrix
#     rw_transition, out_degree, in_degree = get_natural_random_walk_matrix(adjacency_matrix)
#
#     # Calculate base communities(ego excluded) and out-degrees
#     adjacent_nodes = np.ndarray(number_of_nodes, dtype=np.ndarray)
#     base_transitions = np.ndarray(number_of_nodes, dtype=np.ndarray)
#     for n in np.arange(number_of_nodes):
#         rw_transition_row = rw_transition.getrow(n)
#         adjacent_nodes[n] = rw_transition_row.indices.astype(np.int64)
#         base_transitions[n] = rw_transition_row.data
#
#     # Calculate restart probability in the case of lazy PageRank
#     alpha = (alpha_effective * (1 - laziness_factor))/(1 - laziness_factor * alpha_effective)
#
#     number_of_operations_fast = np.zeros(number_of_nodes, dtype=np.int32)
#     execution_time_fast = np.zeros(number_of_nodes, dtype=np.float64)
#
#     number_of_operations_lazy = np.zeros(number_of_nodes, dtype=np.int32)
#     execution_time_lazy = np.zeros(number_of_nodes, dtype=np.float64)
#
#     number_of_operations_rct = np.zeros(number_of_nodes, dtype=np.int32)
#     execution_time_rct = np.zeros(number_of_nodes, dtype=np.float64)
#     s = np.zeros(number_of_nodes, dtype=np.float64)
#     r = np.zeros(number_of_nodes, dtype=np.float64)
#     for n in np.arange(number_of_nodes):
#         print(n)
#
#         # Perform fast PageRank
#         s[:] = 0.0
#         r[:] = 0.0
#         start_time = time.process_time()
#         nop = csimilarity.fast_approximate_personalized_pagerank(s,
#                                                                  r,
#                                                                  base_transitions,
#                                                                  adjacent_nodes,
#                                                                  out_degree,
#                                                                  in_degree,
#                                                                  n,
#                                                                  alpha_effective,
#                                                                  epsilon)
#         elapsed_time = time.process_time() - start_time
#
#         number_of_operations_fast[n] = nop
#         execution_time_fast[n] = elapsed_time
#
#         # Perform lazy PageRank
#         s[:] = 0.0
#         r[:] = 0.0
#         start_time = time.process_time()
#         nop = csimilarity.lazy_approximate_personalized_pagerank(s,
#                                                                  r,
#                                                                  base_transitions,
#                                                                  adjacent_nodes,
#                                                                  out_degree,
#                                                                  in_degree,
#                                                                  n,
#                                                                  alpha,
#                                                                  epsilon)
#         elapsed_time = time.process_time() - start_time
#
#         number_of_operations_lazy[n] = nop
#         execution_time_lazy[n] = elapsed_time
#
#         # Perform Regularized Commute-Time
#         s[:] = 0.0
#         r[:] = 0.0
#         start_time = time.process_time()
#         nop = csimilarity.fast_approximate_regularized_commute(s,
#                                                                r,
#                                                                base_transitions,
#                                                                adjacent_nodes,
#                                                                out_degree,
#                                                                in_degree,
#                                                                n,
#                                                                alpha_effective,
#                                                                epsilon)
#         elapsed_time = time.process_time() - start_time
#
#         number_of_operations_rct[n] = nop
#         execution_time_rct[n] = elapsed_time
#
#     return number_of_operations_fast, execution_time_fast, number_of_operations_lazy, execution_time_lazy, number_of_operations_rct, execution_time_rct
#
#
# def cython_optimization_benchmark(adjacency_matrix, alpha, epsilon):
#     """
#     Compares the efficiency of approaches calculating similarity matrix slices.
#     """
#     adjacency_matrix = adjacency_matrix.tocsr()
#     number_of_nodes = adjacency_matrix.shape[0]
#
#     # Calculate random walk transition probability matrix
#     rw_transition, out_degree, in_degree = get_natural_random_walk_matrix(adjacency_matrix)
#
#     # Calculate base communities(ego excluded) and out-degrees
#     adjacent_nodes = np.ndarray(number_of_nodes, dtype=np.ndarray)
#     base_transitions = np.ndarray(number_of_nodes, dtype=np.ndarray)
#     for n in np.arange(number_of_nodes):
#         rw_transition_row = rw_transition.getrow(n)
#         adjacent_nodes[n] = rw_transition_row.indices.astype(np.int64)
#         base_transitions[n] = rw_transition_row.data
#
#     execution_time_non_opt = np.zeros(number_of_nodes, dtype=np.float64)
#
#     execution_time_opt = np.zeros(number_of_nodes, dtype=np.float64)
#
#     s = np.zeros(number_of_nodes, dtype=np.float64)
#     r = np.zeros(number_of_nodes, dtype=np.float64)
#     for n in np.arange(number_of_nodes):
#         print(n)
#         # Perform cython-optimized absorbing cumulative random walk probability.
#         s[:] = 0.0
#         r[:] = 0.0
#         start_time = time.process_time()
#         nop = csimilarity.fast_approximate_regularized_commute(s,
#                                                                r,
#                                                                base_transitions,
#                                                                      adjacent_nodes,
#                                                                      out_degree,
#                                                                      in_degree,
#                                                                      n,
#                                                                      alpha,
#                                                                      epsilon)
#         elapsed_time = time.process_time() - start_time
#
#         execution_time_opt[n] = elapsed_time
#
#         # Perform unoptimized absorbing cumulative random walk probability.
#         s[:] = 0.0
#         r[:] = 0.0
#         start_time = time.process_time()
#         nop = similarity.fast_approximate_regularized_commute(s,
#                                                               r,
#                                                               base_transitions,
#                                                                     adjacent_nodes,
#                                                                     out_degree,
#                                                                     in_degree,
#                                                                     n,
#                                                                     alpha,
#                                                                     epsilon)
#         elapsed_time = time.process_time() - start_time
#
#         execution_time_non_opt[n] = elapsed_time
#
#     return execution_time_non_opt, execution_time_opt



__author__ = 'Georgios Rizos (georgerizos@iti.gr)'

import numpy as np
from scipy.sparse import issparse

from reveal_graph_embedding.common import load_pickle, store_pickle


def read_features(method_name, path):

    sparse_feature_method_set = set()
    sparse_feature_method_set.update(["lapple",
                                      "apple",
                                      "arcte",
                                      "basecomm",
                                      "mroc",
                                      "edgecluster",
                                      "louvain",
                                      "oslom",
                                      "bigclam"])

    continuous_feature_method_set = set()
    continuous_feature_method_set.update(["lapeig",
                                          "deepwalk",
                                          "rwmodmax",
                                          "repeig"])

    method_name_lowercase = method_name.lower()
    if method_name_lowercase in sparse_feature_method_set:
        features = read_sparse_features(path)
    elif method_name_lowercase in continuous_feature_method_set:
        features = read_continuous_features(path)
    else:
        print("Invalid method name.")
        raise RuntimeError

    return features


def read_continuous_features(path):

    features = np.load(path + ".npy")

    return features


def read_sparse_features(path):

    features = load_pickle(path + ".pkl")
    features = features.tocsr()

    return features


def write_features(path, data):

    if issparse(data):
        write_sparse_data(path, data)
    else:
        write_continuous_features(path, data)


def write_continuous_features(path, data):

    np.save(path, data)


def write_sparse_data(path, data):

    store_pickle(path + ".pkl", data)



__author__ = 'Georgios Rizos (georgerizos@iti.gr)'

import numpy as np
from scipy.sparse import issparse
from sklearn.preprocessing import normalize
from sklearn.utils.validation import check_array
from sklearn.utils.extmath import safe_sparse_dot
from sklearn.preprocessing import LabelBinarizer


def chi2_contingency_matrix(X_train, y_train):
    X = X_train.copy()
    X.data = np.ones_like(X.data)

    X = check_array(X, accept_sparse='csr')
    if np.any((X.data if issparse(X) else X) < 0):
        raise ValueError("Input X must be non-negative.")

    Y = LabelBinarizer().fit_transform(y_train)
    if Y.shape[1] == 1:
        Y = np.append(1 - Y, Y, axis=1)

    observed = safe_sparse_dot(Y.T, X)  # n_classes * n_features

    # feature_count = check_array(X.sum(axis=0))
    # class_prob = check_array(Y.mean(axis=0))
    feature_count = X.sum(axis=0).reshape(1, -1)
    class_prob = Y.mean(axis=0).reshape(1, -1)
    expected = np.dot(class_prob.T, feature_count)

    observed = np.asarray(observed, dtype=np.float64)

    k = len(observed)
    # Reuse observed for chi-squared statistics
    contingency_matrix = observed
    contingency_matrix -= expected
    contingency_matrix **= 2

    expected[expected == 0.0] = 1.0

    contingency_matrix /= expected

    # weights = contingency_matrix.max(axis=0)

    return contingency_matrix


def peak_snr_weight_aggregation(contingency_matrix):
    contingency_matrix[np.isnan(contingency_matrix)] = 0.0

    variance = np.zeros(contingency_matrix.shape[0], dtype=np.float64)
    for k in range(contingency_matrix.shape[0]):
        within_class_signal = contingency_matrix[k, :]
        variance[k] = np.var(within_class_signal)
    variance = np.sqrt(np.mean(variance))

    weights = np.zeros(contingency_matrix.shape[1], dtype=np.float64)
    for f in range(contingency_matrix.shape[1]):
        distribution = contingency_matrix[:, f]
        distribution = distribution[distribution > 0.0]
        if distribution.size > 1:
            signal_to_noise_ratio = (np.max(distribution) - np.min(distribution))/variance
        elif distribution.size == 1:
            signal_to_noise_ratio = np.max(distribution)/variance
        else:
            signal_to_noise_ratio = 0.0
        weights[f] = signal_to_noise_ratio

    # weights = np.zeros(chisq.shape[1], dtype=np.float64)
    # for f in range(chisq.shape[1]):
    #     distribution = chisq[:, f]
    #     distribution = distribution[distribution > 0.0]
    #     if distribution.size > 1:
    #         max_el = np.sum(distribution)
    #         distribution = distribution/np.sum(chisq[:, f])
    #         signal_to_noise_ratio = max_el*((- np.log(1/chisq.shape[0]))/(- np.sum(np.multiply(distribution, np.log(distribution)))))
    #     elif distribution.size == 1:
    #         signal_to_noise_ratio = np.max(distribution)
    #     else:
    #         signal_to_noise_ratio = 0.0
    #     # print(signal_to_noise_ratio)
    #     weights[f] = signal_to_noise_ratio

    return weights


def community_weighting(X_train, X_test, community_weights):
    if issparse(X_train):
        # chi2score = chi2_contingency_matrix(X_train, y_train)
        # chi2score[np.isnan(chi2score)] = 0.0

        X_train = X_train.tocsc()
        X_test = X_test.tocsc()
        for j in range(X_train.shape[1]):
            document_frequency = X_train.getcol(j).data.size
            if document_frequency > 1:
                if community_weights[j] == 0.0:
                    reinforcement = 0.0
                else:
                    reinforcement = np.log(1.0 + community_weights[j])

                X_train.data[X_train.indptr[j]: X_train.indptr[j + 1]] =\
                    X_train.data[X_train.indptr[j]: X_train.indptr[j + 1]]*reinforcement

            document_frequency = X_test.getcol(j).data.size
            if document_frequency > 1:
                if community_weights[j] == 0.0:
                    reinforcement = 0.0
                else:
                    reinforcement = np.log(1.0 + community_weights[j])
                X_test.data[X_test.indptr[j]: X_test.indptr[j + 1]] =\
                    X_test.data[X_test.indptr[j]: X_test.indptr[j + 1]]*reinforcement

        X_train = X_train.tocsr()
        X_test = X_test.tocsr()

        X_train.eliminate_zeros()
        X_test.eliminate_zeros()

        X_train = normalize(X_train, norm="l2")
        X_test = normalize(X_test, norm="l2")
    else:
        pass

    return X_train, X_test


def chi2_psnr_community_weighting(X_train, X_test, y_train):
    if issparse(X_train):
        contingency_matrix = chi2_contingency_matrix(X_train, y_train)
        community_weights = peak_snr_weight_aggregation(contingency_matrix)
        X_train, X_test = community_weighting(X_train, X_test, community_weights)
    else:
        pass

    return X_train, X_test



__author__ = 'Georgios Rizos (georgerizos@iti.gr)'
# cython: profile=True

import numpy as np
cimport numpy as np
cimport cython

FLOAT64 = np.float64
ctypedef np.float64_t FLOAT64_t

INT64 = np.int64
ctypedef np.int64_t INT64_t


@cython.boundscheck(False)
@cython.wraparound(False)
@cython.nonecheck(False)
@cython.cdivision(True)
def pagerank_limit_push(np.ndarray[FLOAT64_t, ndim=1] s,
                        np.ndarray[FLOAT64_t, ndim=1] r,
                        np.ndarray[FLOAT64_t, ndim=1] w_i,
                        np.ndarray[INT64_t, ndim=1] a_i,
                        long push_node,
                        double rho):
    """
    Performs a random step without a self-loop.
    """
    # Calculate the A and B quantities to infinity
    cdef double A_inf = rho*r[push_node]
    cdef double B_inf = (1-rho)*r[push_node]

    # Update approximate Pagerank and residual vectors
    s[push_node] += A_inf
    r[push_node] = 0.0

    cdef long i = 0
    cdef long out_degree = a_i.size

    # Update residual vector at push node's adjacent nodes
    for i in range(out_degree):
        r[a_i[i]] += B_inf * w_i[i]


@cython.boundscheck(False)
@cython.wraparound(False)
@cython.nonecheck(False)
@cython.cdivision(True)
def pagerank_lazy_push(np.ndarray[FLOAT64_t, ndim=1] s,
                       np.ndarray[FLOAT64_t, ndim=1] r,
                       np.ndarray[FLOAT64_t, ndim=1] w_i,
                       np.ndarray[INT64_t, ndim=1] a_i,
                       long push_node,
                       double rho,
                       double laziness_factor):
    """
    Performs a random step with a self-loop.

    Introduced in: Andersen, R., Chung, F., & Lang, K. (2006, October).
                   Local graph partitioning using pagerank vectors.
                   In Foundations of Computer Science, 2006. FOCS'06. 47th Annual IEEE Symposium on (pp. 475-486). IEEE.
    """
    # Calculate the A, B and C quantities
    cdef double A = rho*r[push_node]
    cdef double B = (1-rho)*(1 - laziness_factor)*r[push_node]
    cdef double C = (1-rho)*laziness_factor*(r[push_node])

    # Update approximate Pagerank and residual vectors
    s[push_node] += A
    r[push_node] = C

    cdef long i = 0
    cdef long out_degree = a_i.size

    # Update residual vector at push node's adjacent nodes
    for i in range(out_degree):
        r[a_i[i]] += B * w_i[i]


@cython.boundscheck(False)
@cython.wraparound(False)
@cython.nonecheck(False)
@cython.cdivision(True)
def cumulative_pagerank_difference_limit_push(np.ndarray[FLOAT64_t, ndim=1] s, np.ndarray[FLOAT64_t, ndim=1] r, np.ndarray[FLOAT64_t, ndim=1] w_i, np.ndarray[INT64_t, ndim=1] a_i, long push_node, double rho):
    """
    Performs a random step without a self-loop.
    """
    # Calculate the B quantity to infinity
    cdef double B_inf = (1-rho)*r[push_node]

    # Update approximate regularized commute and residual vectors
    r[push_node] = 0.0

    cdef long i = 0
    cdef long out_degree = a_i.size

    cdef double commute_probability

    # Update residual vector at push node's adjacent nodes
    for i in range(out_degree):
        commute_probability = B_inf * w_i[i]
        s[a_i[i]] += commute_probability
        r[a_i[i]] += commute_probability



__author__ = 'Georgios Rizos (georgerizos@iti.gr)'

import multiprocessing as mp
import itertools
import numpy as np
import scipy.sparse as sparse

from reveal_graph_embedding.common import get_threads_number
from reveal_graph_embedding.eps_randomwalk.transition import get_natural_random_walk_matrix
from reveal_graph_embedding.eps_randomwalk.similarity import fast_approximate_cumulative_pagerank_difference,\
    fast_approximate_personalized_pagerank, lazy_approximate_personalized_pagerank


def parallel_chunks(l, n):
    for thread_id in range(n):
        yield roundrobin_chunks(l, n, thread_id)


def roundrobin_chunks(l, n, id):
    l_c = iter(l)
    x = list(itertools.islice(l_c, id, None, n))
    if len(x):
        return x


def calculate_epsilon_effective(rho, epsilon, seed_degree, neighbor_degrees, mean_degree):
    """
    Semi-automatic effective epsilon threshold calculation.
    """
    # Calculate a weighted neighborhood degree average.
    # neighborhood_degree = rho*seed_degree + (1-rho)*neighbor_degrees.mean()
    neighborhood_degree = neighbor_degrees.mean()

    # Calculate the seed neighborhood normalized effective epsilon.
    epsilon_effective = (epsilon*np.log(1 + seed_degree))/np.log(1 + neighborhood_degree)

    # Calculate the maximum epsilon for at least one push on a neighboring node.
    # Also the minimum epsilon for a push on all the neighboring nodes.
    epsilon_effective_maximum = np.max(1/(seed_degree*neighbor_degrees))
    epsilon_effective_minimum = np.min(1/(seed_degree*neighbor_degrees))

    # print(epsilon_effective, epsilon_effective_maximum, epsilon_effective_minimum)

    # The maximum epsilon is absolute, whereas we regularize for the minimum.
    if epsilon_effective > epsilon_effective_maximum:
        epsilon_effective = epsilon_effective_maximum
    elif epsilon_effective < epsilon_effective_minimum:
        epsilon_effective = (epsilon_effective_minimum + epsilon_effective)/2

    return epsilon_effective


def arcte_with_lazy_pagerank_worker(iterate_nodes,
                                    indices_c,
                                    indptr_c,
                                    data_c,
                                    out_degree,
                                    in_degree,
                                    rho,
                                    epsilon):

    iterate_nodes = np.array(iterate_nodes, dtype=np.int64)
    number_of_nodes = out_degree.size

    mean_degree = np.mean(out_degree)

    rw_transition = sparse.csr_matrix((data_c, indices_c, indptr_c))

    # Store adjacent nodes and corresponding transition weights in array of arrays form.
    adjacent_nodes = np.ndarray(number_of_nodes, dtype=np.ndarray)
    base_transitions = np.ndarray(number_of_nodes, dtype=np.ndarray)
    for n in range(number_of_nodes):
        adjacent_nodes[n] = rw_transition.indices[rw_transition.indptr[n]: rw_transition.indptr[n + 1]]
        base_transitions[n] = rw_transition.data[rw_transition.indptr[n]: rw_transition.indptr[n + 1]]

    # Separate nodes to iterate in smaller chunks in order to compress the feature batches.
    features = sparse.dok_matrix((number_of_nodes, number_of_nodes), dtype=np.float64)
    features = sparse.csr_matrix(features)

    if iterate_nodes.size > 2000:
        node_chunks = list(parallel_chunks(iterate_nodes, iterate_nodes.size//2000))
    else:
        node_chunks = list()
        node_chunks.append(iterate_nodes)

    for node_chunk in node_chunks:
        node_chunk = np.array(node_chunk, dtype=np.int64)

        # Calculate local communities for all nodes.
        row_list = list()
        col_list = list()
        extend_row = row_list.extend
        extend_col = col_list.extend

        # number_of_local_communities = 0

        s = np.zeros(number_of_nodes, dtype=np.float64)  #TODO: What if it is only one?
        r = np.zeros(number_of_nodes, dtype=np.float64)
        for n_index in range(node_chunk.size):
            print(n_index)
            n = node_chunk[n_index]

            # Calculate similarity matrix slice.
            s[:] = 0.0
            r[:] = 0.0

            epsilon_eff = calculate_epsilon_effective(rho, epsilon, out_degree[n], out_degree[adjacent_nodes[n]], mean_degree)

            lazy_rho = (rho*(0.5))/(1-(0.5*rho))

            nop = lazy_approximate_personalized_pagerank(s,
                                                         r,
                                                         base_transitions[:],
                                                         adjacent_nodes[:],
                                                         out_degree,
                                                         in_degree,
                                                         n,
                                                         lazy_rho,
                                                         epsilon_eff)

            s_sparse = sparse.csr_matrix(s)

            # Perform degree normalization of approximate similarity matrix slice.
            relevant_degrees = in_degree[s_sparse.indices]
            s_sparse.data = np.divide(s_sparse.data, relevant_degrees)

            base_community = np.append(adjacent_nodes[n], n)

            # If base community is not strictly non zero, then we break.
            intersection = np.intersect1d(base_community, s_sparse.indices)

            if intersection.size < base_community.size:
                continue

            base_community_rankings = np.searchsorted(s_sparse.indices, base_community)
            min_similarity = np.min(s_sparse.data[base_community_rankings])

            # Sort the degree normalized approximate similarity matrix slice.
            sorted_indices = np.argsort(s_sparse.data, axis=0)
            s_sparse.data = s_sparse.data[sorted_indices]
            s_sparse.indices = s_sparse.indices[sorted_indices]

            most_unlikely_index = s_sparse.indices.size - np.searchsorted(s_sparse.data, min_similarity)

            # Save feature matrix coordinates.
            if most_unlikely_index > base_community.size:
                # print(n_index, out_degree[n], epsilon_eff)
                new_rows = s_sparse.indices[-1:-most_unlikely_index-1:-1]
                extend_row(new_rows)
                # extend_col(number_of_local_communities*np.ones_like(new_rows))
                # number_of_local_communities += 1
                extend_col(n*np.ones_like(new_rows))

        # Form local community feature matrix.
        row = np.array(row_list, dtype=np.int64)
        col = np.array(col_list, dtype=np.int64)
        data = np.ones_like(row, dtype=np.float64)
        # features = sparse.coo_matrix((data, (row, col)), shape=(number_of_nodes, number_of_local_communities))
        chunk_features = sparse.coo_matrix((data, (row, col)), shape=(number_of_nodes, number_of_nodes))
        chunk_features = sparse.csr_matrix(chunk_features)

        features += chunk_features

    return features


def arcte_with_pagerank_worker(iterate_nodes,
                               indices_c,
                               indptr_c,
                               data_c,
                               out_degree,
                               in_degree,
                               rho,
                               epsilon):

    iterate_nodes = np.array(iterate_nodes, dtype=np.int64)
    number_of_nodes = out_degree.size

    mean_degree = np.mean(out_degree)

    rw_transition = sparse.csr_matrix((data_c, indices_c, indptr_c))

    # Store adjacent nodes and corresponding transition weights in array of arrays form.
    adjacent_nodes = np.ndarray(number_of_nodes, dtype=np.ndarray)
    base_transitions = np.ndarray(number_of_nodes, dtype=np.ndarray)
    for n in range(number_of_nodes):
        adjacent_nodes[n] = rw_transition.indices[rw_transition.indptr[n]: rw_transition.indptr[n + 1]]
        base_transitions[n] = rw_transition.data[rw_transition.indptr[n]: rw_transition.indptr[n + 1]]

    # Separate nodes to iterate in smaller chunks in order to compress the feature batches.
    features = sparse.dok_matrix((number_of_nodes, number_of_nodes), dtype=np.float64)
    features = sparse.csr_matrix(features)

    if iterate_nodes.size > 2000:
        node_chunks = list(parallel_chunks(iterate_nodes, iterate_nodes.size//2000))
    else:
        node_chunks = list()
        node_chunks.append(iterate_nodes)

    for node_chunk in node_chunks:
        node_chunk = np.array(node_chunk, dtype=np.int64)

        # Calculate local communities for all nodes.
        row_list = list()
        col_list = list()
        extend_row = row_list.extend
        extend_col = col_list.extend

        # number_of_local_communities = 0

        s = np.zeros(number_of_nodes, dtype=np.float64)  #TODO: What if it is only one?
        r = np.zeros(number_of_nodes, dtype=np.float64)
        for n_index in range(node_chunk.size):
            print(n_index)
            n = node_chunk[n_index]

            # Calculate similarity matrix slice.
            s[:] = 0.0
            r[:] = 0.0

            epsilon_eff = calculate_epsilon_effective(rho, epsilon, out_degree[n], out_degree[adjacent_nodes[n]], mean_degree)

            nop = fast_approximate_personalized_pagerank(s,
                                                         r,
                                                         base_transitions[:],
                                                         adjacent_nodes[:],
                                                         out_degree,
                                                         in_degree,
                                                         n,
                                                         rho,
                                                         epsilon_eff)

            s_sparse = sparse.csr_matrix(s)

            # Perform degree normalization of approximate similarity matrix slice.
            relevant_degrees = in_degree[s_sparse.indices]
            s_sparse.data = np.divide(s_sparse.data, relevant_degrees)

            base_community = np.append(adjacent_nodes[n], n)

            # If base community is not strictly non zero, then we break.
            intersection = np.intersect1d(base_community, s_sparse.indices)

            if intersection.size < base_community.size:
                continue

            base_community_rankings = np.searchsorted(s_sparse.indices, base_community)
            min_similarity = np.min(s_sparse.data[base_community_rankings])

            # Sort the degree normalized approximate similarity matrix slice.
            sorted_indices = np.argsort(s_sparse.data, axis=0)
            s_sparse.data = s_sparse.data[sorted_indices]
            s_sparse.indices = s_sparse.indices[sorted_indices]

            most_unlikely_index = s_sparse.indices.size - np.searchsorted(s_sparse.data, min_similarity)

            # Save feature matrix coordinates.
            if most_unlikely_index > base_community.size:
                # print(n_index, out_degree[n], epsilon_eff)
                new_rows = s_sparse.indices[-1:-most_unlikely_index-1:-1]
                extend_row(new_rows)
                # extend_col(number_of_local_communities*np.ones_like(new_rows))
                # number_of_local_communities += 1
                extend_col(n*np.ones_like(new_rows))

        # Form local community feature matrix.
        row = np.array(row_list, dtype=np.int64)
        col = np.array(col_list, dtype=np.int64)
        data = np.ones_like(row, dtype=np.float64)
        # features = sparse.coo_matrix((data, (row, col)), shape=(number_of_nodes, number_of_local_communities))
        chunk_features = sparse.coo_matrix((data, (row, col)), shape=(number_of_nodes, number_of_nodes))
        chunk_features = sparse.csr_matrix(chunk_features)

        features += chunk_features

    return features


def arcte_worker(iterate_nodes,
                 indices_c,
                 indptr_c,
                 data_c,
                 out_degree,
                 in_degree,
                 rho,
                 epsilon):

    iterate_nodes = np.array(iterate_nodes, dtype=np.int64)
    number_of_nodes = out_degree.size

    mean_degree = np.mean(out_degree)

    rw_transition = sparse.csr_matrix((data_c, indices_c, indptr_c))

    # Store adjacent nodes and corresponding transition weights in array of arrays form.
    adjacent_nodes = np.ndarray(number_of_nodes, dtype=np.ndarray)
    base_transitions = np.ndarray(number_of_nodes, dtype=np.ndarray)
    for n in range(number_of_nodes):
        adjacent_nodes[n] = rw_transition.indices[rw_transition.indptr[n]: rw_transition.indptr[n + 1]]
        base_transitions[n] = rw_transition.data[rw_transition.indptr[n]: rw_transition.indptr[n + 1]]

    # Separate nodes to iterate in smaller chunks in order to compress the feature batches.
    features = sparse.dok_matrix((number_of_nodes, number_of_nodes), dtype=np.float64)
    features = sparse.csr_matrix(features)

    if iterate_nodes.size > 2000:
        node_chunks = list(parallel_chunks(iterate_nodes, iterate_nodes.size//2000))
    else:
        node_chunks = list()
        node_chunks.append(iterate_nodes)

    for node_chunk in node_chunks:
        node_chunk = np.array(node_chunk, dtype=np.int64)

        # Calculate local communities for all nodes.
        row_list = list()
        col_list = list()
        extend_row = row_list.extend
        extend_col = col_list.extend

        # number_of_local_communities = 0

        s = np.zeros(number_of_nodes, dtype=np.float64)  #TODO: What if it is only one?
        r = np.zeros(number_of_nodes, dtype=np.float64)
        for n_index in range(node_chunk.size):
            # print(n_index)
            n = node_chunk[n_index]

            # Calculate similarity matrix slice.
            s[:] = 0.0
            r[:] = 0.0

            epsilon_eff = calculate_epsilon_effective(rho, epsilon, out_degree[n], out_degree[adjacent_nodes[n]], mean_degree)

            nop = fast_approximate_cumulative_pagerank_difference(s,
                                                                  r,
                                                                  base_transitions[:],
                                                                  adjacent_nodes[:],
                                                                  out_degree,
                                                                  in_degree,
                                                                  n,
                                                                  rho,
                                                                  epsilon_eff)

            s_sparse = sparse.csr_matrix(s)

            # Perform degree normalization of approximate similarity matrix slice.
            relevant_degrees = in_degree[s_sparse.indices]
            s_sparse.data = np.divide(s_sparse.data, relevant_degrees)

            base_community = np.append(adjacent_nodes[n], n)
            base_community_rankings = np.searchsorted(s_sparse.indices, base_community)
            min_similarity = np.min(s_sparse.data[base_community_rankings])

            # Sort the degree normalized approximate similarity matrix slice.
            sorted_indices = np.argsort(s_sparse.data, axis=0)
            s_sparse.data = s_sparse.data[sorted_indices]
            s_sparse.indices = s_sparse.indices[sorted_indices]

            most_unlikely_index = s_sparse.indices.size - np.searchsorted(s_sparse.data, min_similarity)

            # Save feature matrix coordinates.
            if most_unlikely_index > base_community.size:
                # print(n_index, out_degree[n], epsilon_eff)
                new_rows = s_sparse.indices[-1:-most_unlikely_index-1:-1]
                extend_row(new_rows)
                # extend_col(number_of_local_communities*np.ones_like(new_rows))
                # number_of_local_communities += 1
                extend_col(n*np.ones_like(new_rows))

        # Form local community feature matrix.
        row = np.array(row_list, dtype=np.int64)
        col = np.array(col_list, dtype=np.int64)
        data = np.ones_like(row, dtype=np.float64)
        # features = sparse.coo_matrix((data, (row, col)), shape=(number_of_nodes, number_of_local_communities))
        chunk_features = sparse.coo_matrix((data, (row, col)), shape=(number_of_nodes, number_of_nodes))
        chunk_features = sparse.csr_matrix(chunk_features)

        features += chunk_features

    return features


def arcte_with_lazy_pagerank(adjacency_matrix, rho, epsilon, number_of_threads=None):
    """
    Extracts local community features for all graph nodes based on the partitioning of node-centric similarity vectors.

    Inputs:  - A in R^(nxn): Adjacency matrix of an undirected network represented as a SciPy Sparse COOrdinate matrix.
             - rho: Restart probability
             - epsilon: Approximation threshold

    Outputs: - X in R^(nxC_n): The latent space embedding represented as a SciPy Sparse COOrdinate matrix.
    """
    adjacency_matrix = sparse.csr_matrix(adjacency_matrix)
    number_of_nodes = adjacency_matrix.shape[0]

    if number_of_threads is None:
        number_of_threads = get_threads_number()
    if number_of_threads == 1:
        # Calculate natural random walk transition probability matrix.
        rw_transition, out_degree, in_degree = get_natural_random_walk_matrix(adjacency_matrix, make_shared=False)

        a = adjacency_matrix.copy()
        a.data = np.ones_like(a.data)
        edge_count_vector = np.squeeze(np.asarray(a.sum(axis=0), dtype=np.int64))

        iterate_nodes = np.where(edge_count_vector != 0)[0]
        argsort_indices = np.argsort(edge_count_vector[iterate_nodes])
        iterate_nodes = iterate_nodes[argsort_indices][::-1]
        iterate_nodes = iterate_nodes[np.where(edge_count_vector[iterate_nodes] > 1.0)[0]]

        # iterate_nodes = np.where(out_degree != 0)[0]
        # argsort_indices = np.argsort(out_degree[iterate_nodes])
        # iterate_nodes = iterate_nodes[argsort_indices][::-1]
        # iterate_nodes = iterate_nodes[np.where(out_degree[iterate_nodes] > 1.0)[0]]

        local_features = arcte_with_lazy_pagerank_worker(iterate_nodes,
                                                         rw_transition.indices,
                                                         rw_transition.indptr,
                                                         rw_transition.data,
                                                         out_degree,
                                                         in_degree,
                                                         rho,
                                                         epsilon)
    else:
        # Calculate natural random walk transition probability matrix.
        rw_transition, out_degree, in_degree = get_natural_random_walk_matrix(adjacency_matrix, make_shared=True)

        a = adjacency_matrix.copy()
        a.data = np.ones_like(a.data)
        edge_count_vector = np.squeeze(np.asarray(a.sum(axis=0), dtype=np.int64))

        iterate_nodes = np.where(edge_count_vector != 0)[0]
        argsort_indices = np.argsort(edge_count_vector[iterate_nodes])
        iterate_nodes = iterate_nodes[argsort_indices][::-1]
        iterate_nodes = iterate_nodes[np.where(edge_count_vector[iterate_nodes] > 1.0)[0]]

        # iterate_nodes = np.where(out_degree != 0)[0]
        # argsort_indices = np.argsort(out_degree[iterate_nodes])
        # iterate_nodes = iterate_nodes[argsort_indices][::-1]
        # iterate_nodes = iterate_nodes[np.where(out_degree[iterate_nodes] > 1.0)[0]]

        pool = mp.Pool(number_of_threads)
        node_chunks = list(parallel_chunks(iterate_nodes, number_of_threads))
        node_count = 0
        for chunk in node_chunks:
            node_count += len(list(chunk))
        results = list()
        for chunk_no in range(len(pool._pool)):
            pool.apply_async(arcte_with_lazy_pagerank_worker,
                             args=(node_chunks[chunk_no],
                                   rw_transition.indices,
                                   rw_transition.indptr,
                                   rw_transition.data,
                                   out_degree,
                                   in_degree,
                                   rho,
                                   epsilon),
                             callback=results.append)
        pool.close()
        pool.join()
        # local_features = sparse.hstack(results)
        local_features = results[0]
        for additive_features in results[1:]:
            local_features += additive_features
        local_features = sparse.csr_matrix(local_features)

    # Form base community feature matrix.
    identity_matrix = sparse.csr_matrix(sparse.eye(number_of_nodes, number_of_nodes, dtype=np.float64))
    adjacency_matrix_ones = adjacency_matrix
    adjacency_matrix_ones.data = np.ones_like(adjacency_matrix.data)
    base_community_features = identity_matrix + adjacency_matrix_ones

    # Stack horizontally matrices to form feature matrix.
    try:
        features = sparse.hstack([base_community_features, local_features]).tocsr()
    except ValueError as e:
        print("Failure with horizontal feature stacking.")
        features = base_community_features

    return features


def arcte_with_pagerank(adjacency_matrix, rho, epsilon, number_of_threads=None):
    """
    Extracts local community features for all graph nodes based on the partitioning of node-centric similarity vectors.

    Inputs:  - A in R^(nxn): Adjacency matrix of an undirected network represented as a SciPy Sparse COOrdinate matrix.
             - rho: Restart probability
             - epsilon: Approximation threshold

    Outputs: - X in R^(nxC_n): The latent space embedding represented as a SciPy Sparse COOrdinate matrix.
    """
    adjacency_matrix = sparse.csr_matrix(adjacency_matrix)
    number_of_nodes = adjacency_matrix.shape[0]

    if number_of_threads is None:
        number_of_threads = get_threads_number()
    if number_of_threads == 1:
        # Calculate natural random walk transition probability matrix.
        rw_transition, out_degree, in_degree = get_natural_random_walk_matrix(adjacency_matrix, make_shared=False)

        a = adjacency_matrix.copy()
        a.data = np.ones_like(a.data)
        edge_count_vector = np.squeeze(np.asarray(a.sum(axis=0), dtype=np.int64))

        iterate_nodes = np.where(edge_count_vector != 0)[0]
        argsort_indices = np.argsort(edge_count_vector[iterate_nodes])
        iterate_nodes = iterate_nodes[argsort_indices][::-1]
        iterate_nodes = iterate_nodes[np.where(edge_count_vector[iterate_nodes] > 1.0)[0]]

        # iterate_nodes = np.where(out_degree != 0)[0]
        # argsort_indices = np.argsort(out_degree[iterate_nodes])
        # iterate_nodes = iterate_nodes[argsort_indices][::-1]
        # iterate_nodes = iterate_nodes[np.where(out_degree[iterate_nodes] > 1.0)[0]]

        local_features = arcte_with_pagerank_worker(iterate_nodes,
                                                    rw_transition.indices,
                                                    rw_transition.indptr,
                                                    rw_transition.data,
                                                    out_degree,
                                                    in_degree,
                                                    rho,
                                                    epsilon)
    else:
        # Calculate natural random walk transition probability matrix.
        rw_transition, out_degree, in_degree = get_natural_random_walk_matrix(adjacency_matrix, make_shared=True)

        a = adjacency_matrix.copy()
        a.data = np.ones_like(a.data)
        edge_count_vector = np.squeeze(np.asarray(a.sum(axis=0), dtype=np.int64))

        iterate_nodes = np.where(edge_count_vector != 0)[0]
        argsort_indices = np.argsort(edge_count_vector[iterate_nodes])
        iterate_nodes = iterate_nodes[argsort_indices][::-1]
        iterate_nodes = iterate_nodes[np.where(edge_count_vector[iterate_nodes] > 1.0)[0]]

        # iterate_nodes = np.where(out_degree != 0)[0]
        # argsort_indices = np.argsort(out_degree[iterate_nodes])
        # iterate_nodes = iterate_nodes[argsort_indices][::-1]
        # iterate_nodes = iterate_nodes[np.where(out_degree[iterate_nodes] > 1.0)[0]]

        pool = mp.Pool(number_of_threads)
        node_chunks = list(parallel_chunks(iterate_nodes, number_of_threads))
        node_count = 0
        for chunk in node_chunks:
            node_count += len(list(chunk))
        results = list()
        for chunk_no in range(len(pool._pool)):
            pool.apply_async(arcte_with_pagerank_worker,
                             args=(node_chunks[chunk_no],
                                   rw_transition.indices,
                                   rw_transition.indptr,
                                   rw_transition.data,
                                   out_degree,
                                   in_degree,
                                   rho,
                                   epsilon),
                             callback=results.append)
        pool.close()
        pool.join()
        # local_features = sparse.hstack(results)
        local_features = results[0]
        for additive_features in results[1:]:
            local_features += additive_features
        local_features = sparse.csr_matrix(local_features)

    # Form base community feature matrix.
    identity_matrix = sparse.csr_matrix(sparse.eye(number_of_nodes, number_of_nodes, dtype=np.float64))
    adjacency_matrix_ones = adjacency_matrix
    adjacency_matrix_ones.data = np.ones_like(adjacency_matrix.data)
    base_community_features = identity_matrix + adjacency_matrix_ones

    # Stack horizontally matrices to form feature matrix.
    try:
        features = sparse.hstack([base_community_features, local_features]).tocsr()
    except ValueError as e:
        print("Failure with horizontal feature stacking.")
        features = base_community_features

    return features


def arcte(adjacency_matrix, rho, epsilon, number_of_threads=None):
    """
    Extracts local community features for all graph nodes based on the partitioning of node-centric similarity vectors.

    Inputs:  - A in R^(nxn): Adjacency matrix of an undirected network represented as a SciPy Sparse COOrdinate matrix.
             - rho: Restart probability
             - epsilon: Approximation threshold

    Outputs: - X in R^(nxC_n): The latent space embedding represented as a SciPy Sparse COOrdinate matrix.
    """
    adjacency_matrix = sparse.csr_matrix(adjacency_matrix)
    number_of_nodes = adjacency_matrix.shape[0]

    if number_of_threads is None:
        number_of_threads = get_threads_number()
    if number_of_threads == 1:
        # Calculate natural random walk transition probability matrix.
        rw_transition, out_degree, in_degree = get_natural_random_walk_matrix(adjacency_matrix, make_shared=False)

        a = adjacency_matrix.copy()
        a.data = np.ones_like(a.data)
        edge_count_vector = np.squeeze(np.asarray(a.sum(axis=0), dtype=np.int64))

        iterate_nodes = np.where(edge_count_vector != 0)[0]
        argsort_indices = np.argsort(edge_count_vector[iterate_nodes])
        iterate_nodes = iterate_nodes[argsort_indices][::-1]
        iterate_nodes = iterate_nodes[np.where(edge_count_vector[iterate_nodes] > 1.0)[0]]

        # iterate_nodes = np.where(out_degree != 0)[0]
        # argsort_indices = np.argsort(out_degree[iterate_nodes])
        # iterate_nodes = iterate_nodes[argsort_indices][::-1]
        # iterate_nodes = iterate_nodes[np.where(out_degree[iterate_nodes] > 1.0)[0]]

        local_features = arcte_worker(iterate_nodes,
                                      rw_transition.indices,
                                      rw_transition.indptr,
                                      rw_transition.data,
                                      out_degree,
                                      in_degree,
                                      rho,
                                      epsilon)
    else:
        # Calculate natural random walk transition probability matrix.
        rw_transition, out_degree, in_degree = get_natural_random_walk_matrix(adjacency_matrix, make_shared=True)

        a = adjacency_matrix.copy()
        a.data = np.ones_like(a.data)
        edge_count_vector = np.squeeze(np.asarray(a.sum(axis=0), dtype=np.int64))

        iterate_nodes = np.where(edge_count_vector != 0)[0]
        argsort_indices = np.argsort(edge_count_vector[iterate_nodes])
        iterate_nodes = iterate_nodes[argsort_indices][::-1]
        iterate_nodes = iterate_nodes[np.where(edge_count_vector[iterate_nodes] > 1.0)[0]]

        # iterate_nodes = np.where(out_degree != 0)[0]
        # argsort_indices = np.argsort(out_degree[iterate_nodes])
        # iterate_nodes = iterate_nodes[argsort_indices][::-1]
        # iterate_nodes = iterate_nodes[np.where(out_degree[iterate_nodes] > 1.0)[0]]

        pool = mp.Pool(number_of_threads)
        node_chunks = list(parallel_chunks(iterate_nodes, number_of_threads))
        node_count = 0
        for chunk in node_chunks:
            node_count += len(list(chunk))
        results = list()
        for chunk_no in range(len(pool._pool)):
            pool.apply_async(arcte_worker,
                             args=(node_chunks[chunk_no],
                                   rw_transition.indices,
                                   rw_transition.indptr,
                                   rw_transition.data,
                                   out_degree,
                                   in_degree,
                                   rho,
                                   epsilon),
                             callback=results.append)
        pool.close()
        pool.join()
        # local_features = sparse.hstack(results)
        local_features = results[0]
        for additive_features in results[1:]:
            local_features += additive_features
        local_features = sparse.csr_matrix(local_features)

    # Form base community feature matrix.
    identity_matrix = sparse.csr_matrix(sparse.eye(number_of_nodes, number_of_nodes, dtype=np.float64))
    adjacency_matrix_ones = adjacency_matrix
    adjacency_matrix_ones.data = np.ones_like(adjacency_matrix.data)
    base_community_features = identity_matrix + adjacency_matrix_ones

    # Stack horizontally matrices to form feature matrix.
    try:
        features = sparse.hstack([base_community_features, local_features]).tocsr()
    except ValueError as e:
        print("Failure with horizontal feature stacking.")
        features = base_community_features

    return features



__author__ = 'Georgios Rizos (georgerizos@iti.gr)'
# cython: profile=True

from collections import deque
import numpy as np
cimport numpy as np
cimport cython
import scipy.sparse as sparse

from reveal_graph_embedding.eps_randomwalk.cython_opt.push import pagerank_limit_push
from reveal_graph_embedding.eps_randomwalk.cython_opt.push import pagerank_lazy_push
from reveal_graph_embedding.eps_randomwalk.cython_opt.push import cumulative_pagerank_difference_limit_push

FLOAT64 = np.float64
ctypedef np.float64_t FLOAT64_t


def fast_approximate_personalized_pagerank(np.ndarray[FLOAT64_t, ndim=1] s,
                                           np.ndarray[FLOAT64_t, ndim=1] r,
                                           np.ndarray w_i,
                                           np.ndarray a_i,
                                           np.ndarray[FLOAT64_t, ndim=1] out_degree,
                                           np.ndarray[FLOAT64_t, ndim=1] in_degree,
                                           long seed_node,
                                           double rho,
                                           double epsilon):
    """
    Calculates the approximate personalized PageRank starting from a seed node without self-loops.
    """
    number_of_nodes = a_i.size

    # Initialize the similarity matrix slice and the residual distribution
    # cdef np.ndarray[FLOAT64_t, ndim=1] s = np.zeros(number_of_nodes, dtype=FLOAT64)
    # cdef np.ndarray[FLOAT64_t, ndim=1] r = np.zeros(number_of_nodes, dtype=FLOAT64)
    r[seed_node] = 1.0

    # Initialize queue of nodes to be pushed
    pushable = deque()
    pushable.append(seed_node)

    # Do one push anyway
    push_node = pushable.popleft()

    pagerank_limit_push(s,
                               r,
                               w_i[push_node],
                               a_i[push_node],
                               push_node,
                               rho)
    cdef long number_of_push_operations = 1

    i = np.where(np.divide(r[a_i[push_node]], in_degree[a_i[push_node]]) >= epsilon)[0]
    if i.size > 0:
        pushable.extend(a_i[push_node][i])

    while len(pushable) > 0:
        # While there are nodes with large residual probabilities, push
        push_node = pushable.popleft()
        if r[push_node]/in_degree[push_node] >= epsilon:
            pagerank_limit_push(s,
                                       r,
                                       w_i[push_node],
                                       a_i[push_node],
                                       push_node,
                                       rho)
            number_of_push_operations += 1

            i = np.where(np.divide(r[a_i[push_node]], in_degree[a_i[push_node]]) >= epsilon)[0]
            if i.size > 0:
                pushable.extend(a_i[push_node][i])

    # Sparsify and return.
    # s_sparse = sparse.csr_matrix(s, shape=(1, number_of_nodes))
    # r_sparse = sparse.csr_matrix(r, shape=(1, number_of_nodes))

    return number_of_push_operations


def lazy_approximate_personalized_pagerank(np.ndarray[FLOAT64_t, ndim=1] s,
                                           np.ndarray[FLOAT64_t, ndim=1] r,
                                           np.ndarray w_i,
                                         np.ndarray a_i,
                                         np.ndarray[FLOAT64_t, ndim=1] out_degree,
                                         np.ndarray[FLOAT64_t, ndim=1] in_degree,
                                         long seed_node,
                                         double rho,
                                         double epsilon,
                                         double laziness_factor=0.5):
    """
    Calculates the approximate personalized PageRank starting from a seed node with self-loops.

    Introduced in: Andersen, R., Chung, F., & Lang, K. (2006, October).
                   Local graph partitioning using pagerank vectors.
                   In Foundations of Computer Science, 2006. FOCS'06. 47th Annual IEEE Symposium on (pp. 475-486). IEEE.
    """
    number_of_nodes = a_i.size

    # Initialize the similarity matrix slice and the residual distribution
    # cdef np.ndarray[FLOAT64_t, ndim=1] s = np.zeros(number_of_nodes, dtype=FLOAT64)
    # cdef np.ndarray[FLOAT64_t, ndim=1] r = np.zeros(number_of_nodes, dtype=FLOAT64)
    r[seed_node] = 1.0

    # Initialize queue of nodes to be pushed
    pushable = deque()
    pushable.append(seed_node)

    # Do one push anyway
    push_node = pushable.popleft()

    pagerank_lazy_push(s,
                              r,
                              w_i[push_node],
                              a_i[push_node],
                              push_node,
                              rho,
                              laziness_factor)
    cdef long number_of_push_operations = 1

    i = np.where(np.divide(r[a_i[push_node]], in_degree[a_i[push_node]]) >= epsilon)[0]
    if i.size > 0:
        pushable.extend(a_i[push_node][i])

    while r[push_node]/in_degree[push_node] >= epsilon:
        pagerank_lazy_push(s,
                                  r,
                                  w_i[push_node],
                                  a_i[push_node],
                                  push_node,
                                  rho,
                                  laziness_factor)
        number_of_push_operations += 1

    # While there are nodes with large residual probabilities, push
    while len(pushable) > 0:
        push_node = pushable.popleft()

        if r[push_node]/in_degree[push_node] >= epsilon:
            pagerank_lazy_push(s,
                                      r,
                                      w_i[push_node],
                                      a_i[push_node],
                                      push_node,
                                      rho,
                                      laziness_factor)
            number_of_push_operations += 1

            i = np.where(np.divide(r[a_i[push_node]], in_degree[a_i[push_node]]) >= epsilon)[0]
            if i.size > 0:
                pushable.extend(a_i[push_node][i])

        while r[push_node]/in_degree[push_node] >= epsilon:
            pagerank_lazy_push(s,
                                      r,
                                      w_i[push_node],
                                      a_i[push_node],
                                      push_node,
                                      rho,
                                      laziness_factor)
            number_of_push_operations += 1

    # Sparsify and return.
    # s_sparse = sparse.csr_matrix(s, shape=(1, number_of_nodes))
    # r_sparse = sparse.csr_matrix(r, shape=(1, number_of_nodes))

    return number_of_push_operations


def fast_approximate_cumulative_pagerank_difference(np.ndarray[FLOAT64_t, ndim=1] s,
                                                    np.ndarray[FLOAT64_t, ndim=1] r,
                                                    np.ndarray w_i,
                                                    np.ndarray a_i,
                                                    np.ndarray[FLOAT64_t, ndim=1] out_degree,
                                                    np.ndarray[FLOAT64_t, ndim=1] in_degree,
                                                    long seed_node,
                                                    double rho,
                                                    double epsilon):
    """
    Calculates cumulative PageRank difference probability starting from a seed node without self-loops.

    Inputs:  - w_i: A NumPy array of arrays of probability transition weights from the seed nodes to its adjacent nodes.
             - a_i: A NumPy array of arrays of the nodes adjacent to the seed node.
             - out_degree: A NumPy array of node out_degrees.
             - in_degree: A NumPy array of node in_degrees.
             - seed_node: The seed for the node-centric personalized PageRank.
             - rho: The restart probability. Usually set in [0.1, 0.2].
             - epsilon: The error threshold.

    Outputs: - s in 1xn: A sparse vector that contains the approximate absorbing random walk cumulative probabilities.
             - r in 1xn: A sparse vector that contains the residual probability distribution.
             - nop: The number of limit probability push operations performed.
    """
    number_of_nodes = a_i.size

    # Initialize the similarity matrix slice and the residual distribution
    # cdef np.ndarray[FLOAT64_t, ndim=1] s = np.zeros(number_of_nodes, dtype=FLOAT64)
    # cdef np.ndarray[FLOAT64_t, ndim=1] r = np.zeros(number_of_nodes, dtype=FLOAT64)
    s[seed_node] = 1.0
    r[seed_node] = 1.0

    # Initialize double-ended queue of nodes to be pushed
    pushable = deque()
    pushable.append(seed_node)

    # Do one push anyway
    push_node = pushable.popleft()

    cumulative_pagerank_difference_limit_push(s,
                                              r,
                                              w_i[push_node],
                                              a_i[push_node],
                                              push_node,
                                              rho)
    cdef long number_of_push_operations = 1

    i = np.where(np.divide(r[a_i[push_node]], in_degree[a_i[push_node]]) >= epsilon)[0]
    if i.size > 0:
        pushable.extend(a_i[push_node][i])

    # While there are nodes with large residual probabilities, push
    while len(pushable) > 0:
        push_node = pushable.popleft()

        # If the threshold is not satisfied, perform a push operation
        # Both this and the later check are needed, since the pushable queue may contain duplicates.
        if r[push_node]/in_degree[push_node] >= epsilon:
            cumulative_pagerank_difference_limit_push(s,
                                                      r,
                                                      w_i[push_node],
                                                      a_i[push_node],
                                                      push_node,
                                                      rho)
            number_of_push_operations += 1

            # Update pushable double-ended queue
            i = np.where(np.divide(r[a_i[push_node]], in_degree[a_i[push_node]]) >= epsilon)[0]
            if i.size > 0:
                pushable.extend(a_i[push_node][i])

    # Sparsify and return.
    # s_sparse = sparse.csr_matrix(s, shape=(1, number_of_nodes))
    # r_sparse = sparse.csr_matrix(r, shape=(1, number_of_nodes))

    return number_of_push_operations



__author__ = 'Georgios Rizos (georgerizos@iti.gr)'

import numpy as np
import scipy.sparse as spsp

from reveal_user_annotation.common.datarw import get_file_row_generator


def read_oslom_features(oslom_folder, number_of_nodes):

    oslom_path = oslom_folder + "/tp"

    number_of_levels = 0

    while True:
        try:
            with open(oslom_path + str(number_of_levels + 1)):
                number_of_levels += 1
        except EnvironmentError:
            number_of_levels += 1
            print('OSLOM hierarchy level: ', number_of_levels)
            break

    number_of_communities = 0
    with open(oslom_path) as fp:
        lines = fp.readlines()
    for line in lines:
        if line[0] == '#':
            continue
        else:
            number_of_communities += 1

    for i in np.arange(1, number_of_levels):
        with open(oslom_path + str(i)) as fp:
            lines = fp.readlines()
        for line in lines:
            if line[0] == '#':
                continue
            else:
                number_of_communities += 1

    features = spsp.dok_matrix((number_of_nodes, number_of_communities), dtype=np.float64)

    j = 0
    with open(oslom_path) as fp:
        lines = fp.readlines()
    for line in lines:
        if line[0] == '#':
            continue
        else:
            words = line.strip().split(' ')
            for word in words:
                features[int(word) - 1, j] = 1
            j += 1

    for i in np.arange(1, number_of_levels):
        with open(oslom_path + str(i)) as fp:
            lines = fp.readlines()
        for line in lines:
            if line[0] == '#':
                continue
            else:
                words = line.strip().split(' ')
                for word in words:
                    features[int(word) - 1, j] = 1
                j += 1

    features = features.tocoo()

    return features


def read_bigclam_features(path, number_of_nodes):
    with open(path, "r") as fp:
        lines_crisp = fp.readlines()
    number_of_communities = len(lines_crisp)

    # Calculate number of nonzero elements
    nnz = 0
    for c in np.arange(number_of_communities):
        words = lines_crisp[c].strip().split('\t')
        nnz += len(words)

    # Form feature matrices
    data_crisp = np.zeros(nnz, dtype=np.float64)
    row_crisp = np.zeros(nnz, dtype=np.int32)
    col_crisp = np.zeros(nnz, dtype=np.int32)

    nnz = 0
    for c in np.arange(number_of_communities):
        words_crisp = lines_crisp[c].strip().split('\t')

        for i in np.arange(len(words_crisp)):
            data_crisp[nnz] = 1.0
            row_crisp[nnz] = int(words_crisp[i])
            col_crisp[nnz] = c

            nnz += 1

    features = spsp.coo_matrix((data_crisp, (row_crisp, col_crisp)),
                               shape=(number_of_nodes, number_of_communities))

    features = features.tocoo()

    return features


def read_matlab_features(array_paths, number_of_nodes, dimensionality):
    """
    Returns a sparse feature matrix as calculated by a Matlab routine.
    """
    # Read the data array
    file_row_gen = get_file_row_generator(array_paths[0], "\t")
    data = list()
    append_data = data.append
    for file_row in file_row_gen:
        append_data(float(file_row[0]))

    # Read the row array
    file_row_gen = get_file_row_generator(array_paths[1], "\t")
    row = list()
    append_row = row.append
    for file_row in file_row_gen:
        append_row(int(float(file_row[0])))

    # Read the data array
    file_row_gen = get_file_row_generator(array_paths[2], "\t")
    col = list()
    append_col = col.append
    for file_row in file_row_gen:
        append_col(int(float(file_row[0])))

    data = np.array(data).astype(np.float64)
    row = np.array(row).astype(np.int64) - 1  # Due to Matlab numbering
    col = np.array(col).astype(np.int64) - 1  # Due to Matlab numbering

    print(np.max(row), np.min(row))
    print(np.max(col), np.min(col))

    # centroids_new = sparse.coo_matrix((data, (row, col)), shape=(number_of_nodes + 1, k))
    features = spsp.coo_matrix((data, (row, col)), shape=(number_of_nodes, dimensionality))

    return features


def read_deepwalk_features(deepwalk_folder, number_of_nodes=None):
    file_row_gen = get_file_row_generator(deepwalk_folder + "/deepwalk.txt", " ")

    first_row = next(file_row_gen)

    if number_of_nodes is not None:
        features = np.zeros((number_of_nodes, int(first_row[1])), dtype=np.float64)
    else:
        features = np.zeros((int(first_row[0]), int(first_row[1])), dtype=np.float64)

    for file_row in file_row_gen:
        node = int(file_row[0]) - 1
        features[node, :] = np.array([np.float64(coordinate) for coordinate in file_row[1:]])

    return features


def read_dense_separated_value_file(file_path, number_of_nodes, separator=","):

    file_row_gen = get_file_row_generator(file_path=file_path, separator=separator)

    first_file_row = next(file_row_gen)
    number_of_dimensions = len(first_file_row)

    features = np.empty((number_of_nodes, number_of_dimensions), dtype=np.float64)

    file_row_counter = 0
    features[file_row_counter, :] = np.array(first_file_row)

    for file_row in file_row_gen:
        file_row_counter += 1
        features[file_row_counter, :] = np.array(file_row)

    return features



__author__ = 'Georgios Rizos (georgerizos@iti.gr)'

import numpy as np


def conductance(adjacency_matrix, node_array):
    number_of_nodes = adjacency_matrix.shape[0]

    node_array_bar = np.setdiff1d(np.arange(number_of_nodes), node_array)

    submatrix = adjacency_matrix[np.ix_(node_array, node_array)]
    submatrix_bar = adjacency_matrix[np.ix_(node_array_bar, node_array_bar)]

    submatrix_volume = submatrix.getnnz()  # TODO: If empty?
    submatrix_bar_volume = submatrix_bar.getnnz()  # TODO: If empty?
    matrix_volume = adjacency_matrix.getnnz()
    cut_volume = (matrix_volume - submatrix_volume - submatrix_bar_volume)/2

    try:
        cut_conductance = cut_volume/min(submatrix_volume, submatrix_bar_volume)
    except ZeroDivisionError:
        cut_conductance = np.Inf

    return cut_conductance


def conductance_and_clustering_coefficient(adjacency_matrix, node_array, seed_node):
    number_of_nodes = adjacency_matrix.shape[0]

    node_array_bar = np.setdiff1d(np.arange(number_of_nodes), node_array)

    submatrix = adjacency_matrix[np.ix_(node_array, node_array)]
    submatrix_bar = adjacency_matrix[np.ix_(node_array_bar, node_array_bar)]

    submatrix_volume = submatrix.getnnz()  # TODO: If empty?
    submatrix_bar_volume = submatrix_bar.getnnz()  # TODO: If empty?
    matrix_volume = adjacency_matrix.getnnz()
    cut_volume = (matrix_volume - submatrix_volume - submatrix_bar_volume)/2

    cut_conductance = cut_volume/min(submatrix_volume, submatrix_bar_volume)

    new_node_array = np.setdiff1d(node_array, seed_node)
    clustering_coefficient = adjacency_matrix[np.ix_(new_node_array, new_node_array)]
    clustering_coefficient = clustering_coefficient.getnnz()/(new_node_array.size*new_node_array.size)

    return cut_conductance, clustering_coefficient


def fast_conductance(array_of_arrays, node_array, matrix_volume):
    submatrix_volume = 0
    cut_volume = 0

    for node in node_array:
        neighbors = array_of_arrays[node]
        degree = neighbors.size

        common = np.intersect1d(node_array, neighbors).size
        submatrix_volume += common
        cut_volume += degree - common

    submatrix_bar_volume = matrix_volume - submatrix_volume - 2*cut_volume

    try:
        cut_conductance = cut_volume/min(submatrix_volume, submatrix_bar_volume)
    except ZeroDivisionError:
        cut_conductance = np.Inf

    return cut_conductance, cut_volume, submatrix_volume


def incremental_conductance(array_of_arrays, node_array, new_node, cut_volume, submatrix_volume, matrix_volume):
    # TODO: What if I have ones in the diagonal?
    neighbors = array_of_arrays[new_node]
    degree = neighbors.size

    common = np.intersect1d(node_array, neighbors).size
    submatrix_volume += common
    cut_volume += degree - common

    submatrix_bar_volume = matrix_volume - submatrix_volume - 2*cut_volume

    try:
        cut_conductance = cut_volume/min(submatrix_volume, submatrix_bar_volume)
    except ZeroDivisionError:
        cut_conductance = np.Inf

    return cut_conductance, cut_volume, submatrix_volume


def decremental_conductance(array_of_arrays, node_array, new_node, cut_volume, submatrix_volume, matrix_volume):
    # TODO: What if I have ones in the diagonal?
    neighbors = array_of_arrays[new_node]
    degree = neighbors.size

    common = np.intersect1d(node_array, neighbors).size
    submatrix_volume -= common
    cut_volume += common - degree

    submatrix_bar_volume = matrix_volume - submatrix_volume - 2*cut_volume

    try:
        cut_conductance = cut_volume/min(submatrix_volume, submatrix_bar_volume)
    except ZeroDivisionError:
        cut_conductance = np.Inf

    return cut_conductance, cut_volume, submatrix_volume
